from __future__ import annotations

import warnings
from dataclasses import asdict
from dataclasses import dataclass
from math import fabs
from typing import Callable
from typing import Iterator
from typing import NamedTuple
from typing import TypeVar

import numpy as np
import pandas as pd
from scipy.integrate import quad
from scipy.interpolate import interp1d
from scipy.linalg import LinAlgError
from scipy.linalg import lstsq
from scipy.optimize import curve_fit
from scipy.optimize import nnls
from scipy.special import erf  # pylint: disable=no-name-in-module
from scipy.stats import pearsonr

from .logger import logger
from .types.pepxmltypes import PepXMLRunRowRT
from .utils import ensure_pos
from .utils import get_abundance
from .utils import isotopicDistribution
from .utils import labelledAtomicAbundances
from .utils import NAMES
from .utils import peptideFormula
from .utils import PeptideSettings
from .utils import resize
from .utils import theoretical_dist

FAIL_E_CURVE = 0
FAIL_W_CURVE = 1
FAIL_W_QUAD = 2
FAIL_W_H_CURVE = 3
FAIL_E_H_CURVE = 4
FAIL_W_PEARSON = 5
FAIL_E_BOUNDARY = 6
FAIL_E_EICS = 7
FAIL_W_PEARSON2 = 8

FAILS = {
    FAIL_E_CURVE: "envelope curve_fit error",
    FAIL_W_CURVE: "envelope curve_fit warning",
    FAIL_W_QUAD: "envelope quadrature warning",
    FAIL_W_H_CURVE: "heavy curve_fit warning",
    FAIL_E_H_CURVE: "heavy curve_fit error",
    FAIL_W_PEARSON: "pearson correlation warning",
    FAIL_E_BOUNDARY: "no intensities withing boundary",
    FAIL_E_EICS: "can't fit EICS",
    FAIL_W_PEARSON2: "pearson correlation warning",
}

LOG_ERRORS = False

R2 = np.sqrt(2.0).astype(np.float64)
RPI2 = np.sqrt(np.pi / 2.0).astype(np.float64)
NSIGMA = 2.0
DELTA_ERF = (RPI2 * (erf(NSIGMA / R2) - erf(-NSIGMA / R2))).astype(np.float64)
USE_QUAD = False
QUAD_LIMIT = 500


def fails(ival: int) -> str:
    return ", ".join(s for i, s in FAILS.items() if ival & (1 << i))


@dataclass
class Envelope:
    # isotopeRegressions: np.ndarray  # 2-D (x,2)
    adjustedRsq: np.ndarray
    isotopeEnvelopes: np.ndarray  # 1-D
    monoFitParams: np.ndarray  # 1-D
    inv_ratio: float
    monoPeakArea: float
    maxPeakArea: float


@dataclass
class HeavyEnvelope:
    heavyDistribution: np.ndarray  # 1-D
    heavyFitParams: np.ndarray  # 1-D
    heavy_adj_r_squared: float


@dataclass
class LabelledEnvelope:
    labelledEnvelopes: np.ndarray  # 1-D
    theoreticalDist: np.ndarray
    relativeIsotopeAbundance: float
    enrichment: float
    labelEnrichment2: float
    heavyCor: float
    heavyCor2: float
    nnls_residual: float
    totalNNLSWeight: float


@dataclass
class AllEnvelopes(Envelope, LabelledEnvelope):
    fail: int = 0

    @classmethod
    def make_all(
        cls,
        fail: int,
        env: Envelope,
        # heavy: HeavyEnvelope,
        labelled: LabelledEnvelope,
    ) -> AllEnvelopes:
        args = {**asdict(env), **asdict(labelled), "fail": fail}
        return cls(**args)  # type: ignore


def normalIntensityFunction(
    rt: float | np.ndarray,
    mu: float,
    sigma: float,
    k: float,
    baseline: float,
) -> float | np.ndarray:
    return k * np.exp(-0.5 * (rt - mu) ** 2 / sigma**2) + baseline


def normalIntensityFunction2(
    rt: float | np.ndarray,
    mu: float,
    sigma: float,
    k: float,
) -> float | np.ndarray:
    return normalIntensityFunction(rt, mu, sigma, k, 0.0)


class RSquared(NamedTuple):
    r_squared: float
    adj_r_squared: float


def lm_adjust(
    mat: np.ndarray,
    x: np.ndarray,
    y: np.ndarray,
    df: int,
) -> RSquared:
    dx = mat @ x - y
    return adjust(dx, y, df)


def curve_adjust(
    func: Callable[[float | np.ndarray, float, float, float], float | np.ndarray],
    popt: np.ndarray,
    x: np.ndarray,
    y: np.ndarray,
    df: int,
) -> RSquared:
    dx = func(x, *popt) - y
    return adjust(dx, y, df)


# https://en.wikipedia.org/wiki/Coefficient_of_determination
def adjust(dx: np.ndarray, y: np.ndarray, df: int) -> RSquared:
    rss = np.sum(dx**2)
    ymean = np.mean(y)
    tss = np.sum((y - ymean) ** 2)
    if tss > 0.0:
        r_squared = 1.0 - rss / tss
        n = len(y)
        if n > df:
            adj_r_squared = 1.0 - ((n - 1) / (n - df)) * (1.0 - r_squared)
        else:
            adj_r_squared = r_squared
    else:
        r_squared, adj_r_squared = np.nan, np.nan
    return RSquared(r_squared, adj_r_squared)


Result = TypeVar("Result")


def catch_warnings(
    func: Callable[[], Result],
    peptide: str,
    msg: str,
    *,
    quiet: bool = False,
) -> tuple[Result, int]:  # Literal[0,1]
    with warnings.catch_warnings(record=True) as warn:
        warnings.simplefilter("always")
        ret = func()
    if warn and not quiet:
        for w in warn:
            message = str(w.message)
            if "\n" in message:
                message, _ = message.split("\n", 1)

            logger.warning(msg, peptide, message)
    return ret, 1 if warn else 0


def fitEIC(
    rawEIC: np.ndarray,  # float[N,2]
    eicIds: np.ndarray,  # bool[N]
    mono_intensity: np.ndarray,  # float[N,3]
) -> Iterator[list[float]]:
    eic: np.ndarray
    for eic in rawEIC:
        try:
            intensity = eic[eicIds, 1]
            x, _residues, rank, _s = lstsq(mono_intensity, intensity)
            alpha, slope = x
            # if rank < 2:
            #     assert residues.shape == ()
            #     residue = np.nan
            # else:
            #     assert residues.shape == (0,)
            #     residue = float(np.sqrt(residues))
            r = lm_adjust(mono_intensity, x, intensity, df=rank)
            adj_r_squared = r.adj_r_squared
        except LinAlgError as e:
            if LOG_ERRORS:
                logger.error("fitEIC: %s", e)
            adj_r_squared, slope, alpha = np.nan, np.nan, np.nan
        yield [adj_r_squared, slope, alpha]


def fitMonoEnvelope(peptide: str, monoEIC: np.ndarray) -> tuple[float, float] | None:
    from .config import INTERPOLATE_INTENSITY  # pylint: disable=import-outside-toplevel

    monoData: pd.DataFrame
    rt, intensity = monoEIC[:, 0], monoEIC[:, 1]
    if INTERPOLATE_INTENSITY:
        # for fill_value='extrapolate'
        # see https://stackoverflow.com/questions/45429831/valueerror-a-value-in-x-new-is-above-the-interpolation-range-what-other-re

        interpolated = interp1d(
            rt,
            intensity,
            assume_sorted=True,
            axis=0,
            fill_value="extrapolate",
        )

        interpolated_rt = np.arange(rt.min(), rt.max() + 1e-4, step=0.5)
        interpolated_i = interpolated(interpolated_rt)
        monoData = pd.DataFrame(dict(rt=interpolated_rt, int=interpolated_i))
        monoData["int"] = monoData["int"].rolling(5, min_periods=1, center=True).mean()
    else:
        monoData = pd.DataFrame(dict(rt=rt, int=intensity))

    monoData = monoData.dropna(axis="index")

    i = monoData["int"].argmax()
    pos = monoData.iloc[i]

    # returns a smaller sigma than R
    def fit() -> np.ndarray:
        popt, *_ = curve_fit(
            normalIntensityFunction,
            monoData["rt"].to_numpy(),
            monoData["int"].to_numpy(),
            p0=(pos["rt"], 3.0, pos["int"], 50.0),
            # method="lm"
            # maxfev=100,
        )
        return popt

    try:
        popt, fail = catch_warnings(
            fit,
            peptide,
            "fitIsotopeEnvelopes::curve_fit[%s]: %s",
            quiet=not LOG_ERRORS,
        )
        fail = fail << FAIL_W_CURVE

    except (RuntimeError, TypeError) as e:
        if LOG_ERRORS:
            logger.error("fitIsotopeEnvelopes::curve_fit[%s]: %s", peptide, e)
        return None
    popt[1] = fabs(popt[1])  # ensure sigma is positive

    mu, sigma, _kscale, _base = popt
    ret = rt.min(), rt.max()

    rng = ret[1] - ret[0]
    if ret[0] <= mu <= ret[1] and sigma <= 2 * rng:
        return mu + rng / 2, mu + rng / 2
    return None


def fitIsotopeEnvelopes(
    peptide: str,
    rawEIC: np.ndarray,
) -> tuple[Envelope | None, int]:
    from .config import (
        INTERPOLATE_INTENSITY,
    )  # pylint: disable=import-outside-toplevel

    monoEIC: np.ndarray = rawEIC[1]
    rt, intensity = monoEIC[:, 0], monoEIC[:, 1]
    if INTERPOLATE_INTENSITY:
        # for fill_value='extrapolate'
        # see https://stackoverflow.com/questions/45429831/valueerror-a-value-in-x-new-is-above-the-interpolation-range-what-other-re

        interpolated = interp1d(
            rt,
            intensity,
            assume_sorted=True,
            axis=0,
            fill_value="extrapolate",
        )

        interpolated_rt = np.arange(rt.min(), rt.max() + 1e-4, step=0.5)
        interpolated_i = interpolated(interpolated_rt)
        monoData = pd.DataFrame(dict(rt=interpolated_rt, int=interpolated_i))
        monoData["int"] = monoData["int"].rolling(5, min_periods=1, center=True).mean()
    else:
        monoData = pd.DataFrame(dict(rt=rt, int=intensity))

    monoData = monoData.dropna(axis="index")

    i = monoData["int"].argmax()  # pylint: disable=unsubscriptable-object
    pos = monoData.iloc[i]

    # returns a smaller sigma than R
    def fit() -> np.ndarray:
        popt, *_ = curve_fit(
            normalIntensityFunction,
            monoData["rt"].to_numpy(),
            monoData["int"].to_numpy(),
            p0=(pos["rt"], 3.0, pos["int"], 50.0),
            # method="lm"
            # maxfev=100,
        )
        return popt

    try:
        popt, fail = catch_warnings(
            fit,
            peptide,
            "fitIsotopeEnvelopes::curve_fit[%s]: %s",
            quiet=not LOG_ERRORS,
        )
        fail = fail << FAIL_W_CURVE

    except (RuntimeError, TypeError) as e:
        if LOG_ERRORS:
            logger.error("fitIsotopeEnvelopes::curve_fit[%s]: %s", peptide, e)
        return None, 1 << FAIL_E_CURVE
    popt[1] = fabs(popt[1])  # ensure sigma is positive

    mu, sigma, kscale, base = popt

    lBoundary: float = max(mu - NSIGMA * sigma, rt[0])
    rBoundary: float = min(mu + NSIGMA * sigma, rt[-1])

    eicIds = (rt >= lBoundary) & (rt <= rBoundary)

    if eicIds.sum() == 0:
        return None, fail | (1 << FAIL_E_BOUNDARY)

    if USE_QUAD and INTERPOLATE_INTENSITY:

        def integrate() -> float:
            monoPeakArea, _ = quad(
                interpolated,
                lBoundary,
                rBoundary,
                limit=QUAD_LIMIT,
                # epsabs=1e-6,
                epsrel=1e-6,
            )
            return monoPeakArea

        monoPeakArea, fail2 = catch_warnings(
            integrate,
            peptide,
            "fitIsotopeEnvelopes::quad[%s]: %s",
            quiet=not LOG_ERRORS,
        )
        fail |= fail2 << FAIL_W_QUAD
        # if not fail2:
        #     monoPeakArea2 = base * (rBoundary - lBoundary) + kscale * sigma * RPI2 * (
        #         erf((rBoundary - mu) / sigma / R2) - erf((lBoundary - mu) / sigma / R2)
        #     )
        #     a_ = np.abs(monoPeakArea2 - monoPeakArea) / monoPeakArea2

        #     if a_ > 1e-2:
        #         logger.debug(
        #             "can't integrate: %s != %s (%s)", monoPeakArea, monoPeakArea2, a_
        #         )

    else:
        monoPeakArea = base * (rBoundary - lBoundary) + kscale * sigma * RPI2 * (
            erf((rBoundary - mu) / sigma / R2) - erf((lBoundary - mu) / sigma / R2)
        )

    intensity = intensity[eicIds]

    im = np.c_[np.ones(len(intensity), dtype=np.float32), intensity]
    isotopeRegressions = np.array(
        list(fitEIC(rawEIC, eicIds, im)),
        dtype=np.float32,
    )  # float[N=maxIso+2,3] # columns: adj_rsquared, alpha, slope (beta)
    alpha, slopes = isotopeRegressions[:, 2], isotopeRegressions[:, 1]
    anynan = np.isnan(slopes)
    nfailed = np.sum(anynan)
    if nfailed == len(slopes):
        return None, 1 << FAIL_E_EICS

    a = alpha * (rBoundary - lBoundary)
    isotopeEnvelopes = ensure_pos(a + monoPeakArea * slopes)

    maxPeakArea = isotopeEnvelopes[1:].max()
    mono = isotopeEnvelopes[1]
    inv_ratio = isotopeEnvelopes[0] / mono if mono > 0.0 else np.inf
    return (
        Envelope(
            # isotopeRegressions=isotopeRegressions,  # only used for filtering in export
            adjustedRsq=isotopeRegressions[:, 0].astype(np.float32),
            isotopeEnvelopes=isotopeEnvelopes.astype(np.float32),
            # mu, sigma, scale, baseline
            monoFitParams=popt.astype(np.float32),  # for plotting the fitted gaussian
            inv_ratio=inv_ratio,
            monoPeakArea=monoPeakArea,
            maxPeakArea=maxPeakArea,
        ),
        fail,
    )


def heavy_dist(peptide: str, isotopeEnvelopeInfo: Envelope) -> np.ndarray:
    ndist = isotopicDistribution(peptide)

    isotopeEnvelopes = isotopeEnvelopeInfo.isotopeEnvelopes
    monoPeakArea = isotopeEnvelopes[1]

    naturalIsotopeEnvelope = monoPeakArea * ndist / ndist[0]
    naturalIsotopeEnvelope = resize(naturalIsotopeEnvelope, len(isotopeEnvelopes) - 1)

    heavyDistribution = ensure_pos(
        isotopeEnvelopes[1:] - naturalIsotopeEnvelope,
    ).astype(np.float32)
    return heavyDistribution


def fitHeavyEnvelope(
    peptide: str,
    isotopeEnvelopeInfo: Envelope,
) -> tuple[HeavyEnvelope, int]:
    heavyDistribution = heavy_dist(peptide, isotopeEnvelopeInfo)

    nh = len(heavyDistribution)

    idx = np.array(list(range(0, nh)), dtype=np.float32)
    i = heavyDistribution.argmax()
    scale, mu = heavyDistribution[i], idx[i]

    try:

        def fit() -> np.ndarray:
            popt, *_ = curve_fit(
                normalIntensityFunction2,
                idx,
                heavyDistribution,
                p0=(mu, 1.0, scale),
                maxfev=1000,
            )
            return popt

        popt, fail = catch_warnings(
            fit,
            peptide,
            "fitHeavyEnvelope::curve_fit[%s]: %s",
            quiet=not LOG_ERRORS,
        )
        fail = fail << FAIL_W_H_CURVE

    except (RuntimeError, TypeError) as e:
        if LOG_ERRORS:
            logger.error("fitHeavyEnvelope::curve_fit[%s]: %s", peptide, e)
        return (
            HeavyEnvelope(
                heavyDistribution=heavyDistribution,
                heavyFitParams=np.zeros(shape=(5,), dtype=np.float32),
                heavy_adj_r_squared=0.0,
            ),
            1 << FAIL_E_H_CURVE,
        )

    popt[1] = fabs(popt[1])  # ensure sigma is positive

    r = curve_adjust(
        normalIntensityFunction2,
        popt,
        idx,
        heavyDistribution,
        df=len(popt),
    )
    # HEAVYFITPARAMS = ('mu','sigma','scale','r_squared', 'adj_r_squared')
    heavyFitParams = np.array([*popt, r.r_squared, r.adj_r_squared], dtype=np.float32)

    return (
        HeavyEnvelope(
            heavyDistribution=heavyDistribution,
            heavyFitParams=heavyFitParams,
            heavy_adj_r_squared=r.adj_r_squared,
        ),
        fail,
    )


def get_enrichments(peptide: str, settings: PeptideSettings) -> np.ndarray:
    """elementCount + 1 array of enrichments"""
    naturalLabelledElement = get_abundance(settings)
    formula = peptideFormula(peptide)
    elementCount = formula[NAMES == settings.labelledElement][0]

    de = (settings.maximumLabelEnrichment - naturalLabelledElement) / elementCount
    return np.array(
        [naturalLabelledElement + i * de for i in range(elementCount + 1)],
        dtype=np.float32,
    )


def labelledEnvelopeCalculation(
    peptide: str,
    maxIso: int,
    isotopeEnvelopeInfo: Envelope,
    # heavyEnvelopeInfo: HeavyEnvelope,
    settings: PeptideSettings,
) -> tuple[LabelledEnvelope, int]:
    enrichments = get_enrichments(peptide, settings)
    ncols = len(enrichments)
    elementCount = ncols - 1
    isotopeEnvelopeBasis = np.zeros(shape=(maxIso + 1, ncols))
    for i, elementEnrichmentLevel in enumerate(enrichments):
        abundances = labelledAtomicAbundances(
            settings.labelledIsotopeNumber,
            settings.labelledElement,
            elementEnrichmentLevel,
        )

        d = isotopicDistribution(peptide, abundances)
        maxEl = min(len(d), maxIso + 1)
        isotopeEnvelopeBasis[:maxEl, i] = d[:maxEl]

    isotopeEnvelopes = isotopeEnvelopeInfo.isotopeEnvelopes
    isotopeEnvelopes = resize(ensure_pos(isotopeEnvelopes[1:]), maxIso + 1)
    labelledEnvelope: np.ndarray

    # area = isotopeEnvelopeInfo.monoPeakArea

    labelledEnvelope, nnls_residual = nnls(isotopeEnvelopeBasis, isotopeEnvelopes)

    nnls_residual = np.sqrt(nnls_residual) if nnls_residual > 0 else 0.0

    totalNNLSWeight = labelledEnvelope.sum()

    labelEnrichment = 0.0
    labelEnrichment2 = 0.0
    relativeIsotopeAbundance = 0.0

    if totalNNLSWeight > 0:
        # this is the same as LPF (other authors had different definitions of RIA)
        relativeIsotopeAbundance = 1.0 - labelledEnvelope[0] / totalNNLSWeight

        labelEnrichment2 = enrichments @ labelledEnvelope / totalNNLSWeight
        labelEnrichment2 = min(
            max(0, labelEnrichment2),
            settings.maximumLabelEnrichment,
        )

    theoreticalMaxEnrichment = elementCount * labelledEnvelope[1:].sum()
    if theoreticalMaxEnrichment > 0:
        labelEnrichment = (
            (np.array(range(len(labelledEnvelope))) * labelledEnvelope).sum()
            / theoreticalMaxEnrichment
            * settings.maximumLabelEnrichment
        )

    if totalNNLSWeight > 0:
        # x /= totalNNLSWeight
        theoreticalDist = isotopeEnvelopeBasis @ labelledEnvelope
        # nnls_residual /= totalNNLSWeight
    else:
        theoreticalDist = theoretical_dist(peptide, labelEnrichment2, settings)

    heavyDistribution = heavy_dist(peptide, isotopeEnvelopeInfo)
    # heavyDistribution = heavyEnvelopeInfo.heavyDistribution
    nmax = max(len(theoreticalDist), len(heavyDistribution), len(isotopeEnvelopes))
    theoreticalDist = resize(theoreticalDist, nmax)
    heavyDistribution = resize(heavyDistribution, nmax)
    isotopeEnvelopes = resize(isotopeEnvelopes, nmax)

    def fit() -> float:
        res = pearsonr(heavyDistribution, theoreticalDist)
        return res.statistic

    def fit2() -> float:
        res = pearsonr(isotopeEnvelopes, theoreticalDist)
        return res.statistic

    heavyCor, fail = catch_warnings(
        fit,
        peptide,
        "labelledEnvelopeCalculation::pearsonr[%s]: %s",
        quiet=not LOG_ERRORS,
    )
    fail = fail << FAIL_W_PEARSON
    heavyCor2, fail2 = catch_warnings(
        fit2,
        peptide,
        "labelledEnvelopeCalculation::pearsonr2[%s]: %s",
        quiet=not LOG_ERRORS,
    )
    fail |= fail2 << FAIL_W_PEARSON2

    return (
        LabelledEnvelope(
            labelledEnvelopes=labelledEnvelope.astype(np.float32),
            theoreticalDist=theoreticalDist.astype(np.float32),
            relativeIsotopeAbundance=relativeIsotopeAbundance,
            enrichment=labelEnrichment,
            labelEnrichment2=labelEnrichment2,
            heavyCor=heavyCor,
            heavyCor2=heavyCor2,
            nnls_residual=nnls_residual,
            totalNNLSWeight=totalNNLSWeight,
        ),
        fail,
    )


def fitEnvelopes(
    pep: PepXMLRunRowRT,  # only need "peptide" , "maxIso"
    rawEIC: np.ndarray,  # [N==pep.maxIso+2,M,2]
    settings: PeptideSettings,
) -> AllEnvelopes | None:
    """Fit envelopes to raw EIC matrix of [mzranges, rt, intensity]"""
    iso, faile = fitIsotopeEnvelopes(pep.peptide, rawEIC)
    if iso is None:
        return None
    # heavy, failh = fitHeavyEnvelope(pep.peptide, iso)
    labelled, faill = labelledEnvelopeCalculation(
        pep.peptide,
        pep.maxIso,
        iso,
        # heavy,
        settings,
    )
    return AllEnvelopes.make_all(faile | faill, iso, labelled)



from __future__ import annotations

from pathlib import Path
from typing import TypedDict

import pandas as pd  # type: ignore
from sqlalchemy import Boolean
from sqlalchemy import create_engine
from sqlalchemy import Float
from sqlalchemy import ForeignKey
from sqlalchemy import func
from sqlalchemy import inspect
from sqlalchemy import Integer
from sqlalchemy import JSON
from sqlalchemy import select
from sqlalchemy import String
from sqlalchemy import Text
from sqlalchemy.engine import Connectable
from sqlalchemy.engine import Engine
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import object_session
from sqlalchemy.orm import relationship
from sqlalchemy.orm import with_parent

from .utils import check_missing_columns
from .utils import secondary

# func.count()
# pylint: disable=not-callable


class Base(DeclarativeBase):
    pass


def clear_orm(Orm: type) -> None:
    if hasattr(Orm, "__table__"):
        Base.metadata.remove(getattr(Orm, "__table__"))
    elif hasattr(Orm, "__tablename__"):
        tname = getattr(Orm, "__tablename__")
        if tname in Base.metadata.tables:
            Base.metadata.remove(Base.metadata.tables[tname])
    Base.registry.dispose()


class Modification(TypedDict):
    mass: float
    massdiff: float | None
    position: int
    site: str | None


class Peptide(Base):
    __tablename__ = "peptides"
    peptideid: Mapped[int] = mapped_column(Integer, primary_key=True)
    peptide: Mapped[str] = mapped_column(Text, nullable=False)
    assumed_charge: Mapped[int | None] = mapped_column(Integer)
    modcol: Mapped[str | None] = mapped_column(Text)
    calc_neutral_pep_mass: Mapped[float | None] = mapped_column(Float)
    precursor_neutral_mass: Mapped[float | None] = mapped_column(Float)
    peptideprophet_probability: Mapped[float | None] = mapped_column(Float)
    mz: Mapped[float | None] = mapped_column(Float)
    observed_mz: Mapped[float | None] = mapped_column(Float)
    retention_time_sec: Mapped[float | None] = mapped_column(Float)
    num_missed_cleavages: Mapped[int | None] = mapped_column(Integer)
    is_decoy: Mapped[bool | None] = mapped_column(Boolean)
    agg_count: Mapped[int | None] = mapped_column(Integer)
    # proteins: Mapped[list[str]] = mapped_column(JSON)
    protein_descr: Mapped[list[str] | None] = mapped_column(JSON)
    modifications: Mapped[list[Modification] | None] = mapped_column(JSON)
    run: Mapped[str | None] = mapped_column(String(64))
    isLabelledRun: Mapped[bool | None] = mapped_column(Boolean)
    fdr: Mapped[float | None] = mapped_column(Float)
    # probability = Column(Float)
    # percent_coverage = Column(Float)
    # confidence = Column(Float)
    # ngroups = Column(Integer)
    # prStr = Column(Text)
    # description = Column(Text)
    # prGrassumed_chargeoup = Column(Integer)
    # ----- computed -----
    labelledElementCount: Mapped[int | None] = mapped_column(Integer)
    maxIso: Mapped[int | None] = mapped_column(Integer)
    mzranges: Mapped[list[float] | None] = mapped_column(JSON, deferred=True)
    eics: Mapped[list[float] | None] = mapped_column(JSON, deferred=True)
    # isotopeRegressions = Column(JSON)
    adjustedRsq: Mapped[list[float] | None] = mapped_column(
        JSON,
        deferred=True,
    )  # list[float]
    isotopeEnvelopes: Mapped[list[float] | None] = mapped_column(JSON, deferred=True)
    monoPeakArea: Mapped[float | None] = mapped_column(Float)
    maxPeakArea: Mapped[float | None] = mapped_column(Float)
    monoFitParams: Mapped[list[float] | None] = mapped_column(JSON, deferred=True)
    inv_ratio: Mapped[float | None] = mapped_column(Float)
    # heavyDistribution = Column(JSON)
    # heavyFitParams = Column(JSON)
    # heavy_adj_r_squared = Column(Float)
    labelledEnvelopes: Mapped[list[float] | None] = mapped_column(JSON, deferred=True)
    relativeIsotopeAbundance: Mapped[float | None] = mapped_column(Float)
    enrichment: Mapped[float | None] = mapped_column(Float)
    labelEnrichment2: Mapped[float | None] = mapped_column(Float)
    heavyCor: Mapped[float | None] = mapped_column(Float)
    heavyCor2: Mapped[float | None] = mapped_column(Float)
    theoreticalDist: Mapped[list[float] | None] = mapped_column(JSON, deferred=True)
    nnls_residual: Mapped[float | None] = mapped_column(Float)
    totalNNLSWeight: Mapped[float] = mapped_column(Float)
    fail: Mapped[int] = mapped_column(Integer)
    eics_shape: Mapped[list[int] | None] = mapped_column(JSON, deferred=True)
    protein_names: Mapped[list[str] | None] = mapped_column(JSON)

    proteins: Mapped[list[Protein]] = relationship(
        lambda: Protein,
        secondary="peptides_proteins",
        back_populates="peptides",
    )

    @hybrid_property
    def norm_nnls_residual(self) -> float:
        if self.totalNNLSWeight == 0.0:
            return 0.0
        if self.nnls_residual is None:
            return 0.0
        return self.nnls_residual / self.totalNNLSWeight

    @norm_nnls_residual.expression
    def _norm_nnls_residual(cls):  # pylint: disable=no-self-argument
        return cls.nnls_residual / cls.totalNNLSWeight

    @hybrid_property
    def protein_count(self):
        q = select(func.count()).where(with_parent(self, Peptide.proteins))
        return object_session(self).execute(q).scalar_one()

    @protein_count.expression
    def _protein_count_expression(cls):  # pylint: disable=no-self-argument
        q = (
            select(func.count(PepProt.peptideid))
            .where(PepProt.peptideid == cls.peptideid)
            .label("protein_count")
        )
        return q

    def __repr__(self):
        return f"Peptide({self.peptide},{self.mz}, {self.assumed_charge})"

    iso_peaks_nr_20: Mapped[int] = mapped_column(
        Integer,
        nullable=False,
        server_default="0",
    )
    iso_peaks_nr_40: Mapped[int] = mapped_column(
        Integer,
        nullable=False,
        server_default="0",
    )
    iso_peaks_nr_60: Mapped[int] = mapped_column(
        Integer,
        nullable=False,
        server_default="0",
    )
    iso_peaks_nr_80: Mapped[int] = mapped_column(
        Integer,
        nullable=False,
        server_default="0",
    )


MINRSQ100 = [20, 40, 60, 80]


class PI(TypedDict):
    protein_name: str


class PeptideInfo(TypedDict):
    calc_neutral_pep_mass: float
    charge: int
    exp_tot_instances: float
    fpkm_adjusted_probability: float
    group_weight: str  # str of float
    initial_probability: float
    is_contributing_evidence: bool
    is_nondegenerate_evidence: bool
    mod_aminoacid_mass: float | None
    mod_nterm_mass: float | None
    modified_peptide: str | None
    n_enzymatic_termini: int
    n_instances: int
    n_sibling_peptides: float
    n_sibling_peptides_bin: int
    nsp_adjusted_probability: float
    peptide_group_designator: str
    peptide_parent_protein: list[PI] | None
    peptide_sequence: str
    weight: float


class Protein(Base):
    __tablename__ = "proteins"

    proteinid: Mapped[int] = mapped_column(Integer, primary_key=True)
    probability: Mapped[int | None] = mapped_column(Float)
    group_number: Mapped[int | None] = mapped_column(Integer)
    protein_name: Mapped[str | None] = mapped_column(Text)
    n_indistinguishable_proteins: Mapped[int | None] = mapped_column(Integer)
    percent_coverage: Mapped[float | None] = mapped_column(Float)
    unique_stripped_peptides: Mapped[list[str] | None] = mapped_column(JSON)
    group_sibling_id: Mapped[str | None] = mapped_column(String(16))
    total_number_peptides: Mapped[int | None] = mapped_column(Integer)
    total_number_distinct_peptides: Mapped[int | None] = mapped_column(Integer)
    confidence: Mapped[float | None] = mapped_column(Float)
    peptide: Mapped[list[PeptideInfo] | None] = mapped_column(JSON)
    protein_description: Mapped[str | None] = mapped_column(Text)
    pct_spectrum_ids: Mapped[float | None] = mapped_column(Float)
    prot_length: Mapped[int | None] = mapped_column(Integer)
    indistinguishable_protein: Mapped[list[str] | None] = mapped_column(JSON)
    lpf_median: Mapped[float | None] = mapped_column(Float)
    lpf_std: Mapped[float | None] = mapped_column(Float)
    num_found_peptides: Mapped[int | None] = mapped_column(Integer)
    # subsuming_protein_entry = Column(JSON)
    peptides: Mapped[list[Peptide]] = relationship(
        Peptide,
        secondary="peptides_proteins",
        back_populates="proteins",
    )

    @hybrid_property
    def peptide_count(self) -> int:
        sess = object_session(self)
        if sess is None:
            raise RuntimeError("no session for Peptide")
        q = select(func.count()).where(with_parent(self, Protein.peptides))
        return sess.execute(q).scalar_one()

    @peptide_count.expression
    def _peptide_count_expression(cls):  # pylint: disable=no-self-argument
        q = (
            select(func.count())
            .where(PepProt.proteinid == cls.proteinid)
            .label("peptide_count")
        )
        return q

    def __repr__(self) -> str:
        return f"Protein({self.protein_name},prob={self.probability},npep={self.unique_stripped_peptides})"


class PepProt(Base):
    __tablename__ = "peptides_proteins"
    peptideid: Mapped[int] = mapped_column(
        Integer,
        ForeignKey(Peptide.peptideid),
        nullable=False,
        primary_key=True,
    )
    proteinid: Mapped[int] = mapped_column(
        Integer,
        ForeignKey(Protein.proteinid),
        nullable=False,
        primary_key=True,
    )


def aggregate_lpf(
    prot: pd.DataFrame,
    pep: pd.DataFrame,
    secondary: pd.DataFrame,
) -> pd.DataFrame:
    m = secondary.merge(pep, how="inner", left_on="peptideid", right_on="peptideid")
    a = (
        m.groupby("proteinid")
        .agg(
            lpf_median=pd.NamedAgg(column="relativeIsotopeAbundance", aggfunc="median"),
            lpf_std=pd.NamedAgg(column="relativeIsotopeAbundance", aggfunc="std"),
            num_found_peptides=pd.NamedAgg(column="peptideid", aggfunc="count"),
        )
        .reset_index()
    )
    a.fillna(0.0, inplace=True)

    return prot.merge(a, how="left", left_on="proteinid", right_on="proteinid")


def save_db(
    pep: pd.DataFrame,
    prot: pd.DataFrame,
    engine: Engine,
) -> tuple[int, int, int]:
    from .utils import dehydrate_peptides, dehydrate_prot
    from .iso_peaks import add_iso_peaks

    # from .precompute import populate_isoPeaksNrFixed
    Base.metadata.create_all(bind=engine)

    pep = dehydrate_peptides(pep)
    prot = dehydrate_prot(prot)
    # pep["peptideid"] = pep.index + 1
    prot["proteinid"] = prot.index + 1
    sec, names = secondary(pep, prot)
    # add protein_names concatenation to peptide
    pep = pep.merge(names, how="left", left_on="peptideid", right_on="peptideid")
    # add iso_peaks_nrNN peaks
    pep = add_iso_peaks(pep)
    # don't save any proteins that don't exists
    prot = prot[prot["proteinid"].isin(sec["proteinid"])]

    prot = aggregate_lpf(prot, pep, sec)
    with engine.begin() as conn:
        npep = save_df(pep, Peptide, conn)
        nprot = save_df(prot, Protein, conn)
        nsec = save_df(sec, PepProt, conn)

    # populate_isoPeaksNrFixed(engine, force=True)
    return npep, nprot, nsec


def save_df(
    df: pd.DataFrame,
    Orm: type[Base],
    conn: Connectable,
    chunksize: int = 1000,
) -> int:
    i = inspect(Orm)
    dtypes = {c.name: c.type for c in i.columns}

    return df[list(dtypes)].to_sql(
        i.local_table.name,  # type: ignore
        con=conn,
        # need this for some reason....
        dtype=dtypes,
        chunksize=chunksize,
        if_exists="append",
        index=False,
    )


def check_missing_pep(pep: pd.DataFrame) -> set[str]:
    return check_missing_columns(pep, Peptide)


def check_missing_prot(prot: pd.DataFrame) -> set[str]:
    return check_missing_columns(prot, Protein)


def file2engine(filename: str | Path, *, echo: bool = False) -> Engine:
    return create_engine(f"sqlite:///{filename}", future=True, echo=echo)


def save_to_file(
    sqlfile: Path,
    pep: pd.DataFrame,
    prot: pd.DataFrame,
) -> tuple[int, int, int]:
    if sqlfile.exists():
        sqlfile.unlink()

    engine = file2engine(sqlfile)
    try:
        return save_db(pep, prot, engine)
    finally:
        engine.dispose()


if __name__ == "__main__":
    import click

    @click.command()
    @click.argument("database")
    def dbtest(database: str) -> None:
        db = Path(database)
        if db.is_file():
            db.unlink(missing_ok=True)
        click.secho(f"creating {database}")
        engine = file2engine(database)
        Base.metadata.create_all(bind=engine)

    dbtest()  # pylint: disable=no-value-for-parameter



from __future__ import annotations

import numpy as np
from matplotlib.axes import Axes

# culled from seaborn so we can remove one dependency...


def despine(
    ax: Axes,
    *,
    top: bool = True,
    right: bool = True,
    left: bool = False,
    bottom: bool = False,
    offset: int | dict[str, int] | None = None,
    trim: bool = False,
) -> None:
    A = dict(top=top, bottom=bottom, left=left, right=right)
    for side, aval in A.items():
        # Toggle the spine objects
        is_visible = not aval
        ax.spines[side].set_visible(is_visible)
        if offset is not None and is_visible:
            try:
                val = offset.get(side, 0)  # type: ignore
            except AttributeError:
                val = offset
            ax.spines[side].set_position(("outward", val))

    # Potentially move the ticks
    if left and not right:
        maj_on = any(t.tick1line.get_visible() for t in ax.yaxis.majorTicks)
        min_on = any(t.tick1line.get_visible() for t in ax.yaxis.minorTicks)
        ax.yaxis.set_ticks_position("right")
        for t in ax.yaxis.majorTicks:
            t.tick2line.set_visible(maj_on)
        for t in ax.yaxis.minorTicks:
            t.tick2line.set_visible(min_on)

    if bottom and not top:
        maj_on = any(t.tick1line.get_visible() for t in ax.xaxis.majorTicks)
        min_on = any(t.tick1line.get_visible() for t in ax.xaxis.minorTicks)
        ax.xaxis.set_ticks_position("top")
        for t in ax.xaxis.majorTicks:
            t.tick2line.set_visible(maj_on)
        for t in ax.xaxis.minorTicks:
            t.tick2line.set_visible(min_on)

    if trim:
        # clip off the parts of the spines that extend past major ticks
        xticks = np.asarray(ax.get_xticks())
        if xticks.size:
            firsttick = np.compress(xticks >= min(ax.get_xlim()), xticks)[0]
            lasttick = np.compress(xticks <= max(ax.get_xlim()), xticks)[-1]
            ax.spines["bottom"].set_bounds(firsttick, lasttick)
            ax.spines["top"].set_bounds(firsttick, lasttick)
            newticks = xticks.compress(xticks <= lasttick)
            newticks = newticks.compress(newticks >= firsttick)
            ax.set_xticks(newticks)

        yticks = np.asarray(ax.get_yticks())
        if yticks.size:
            firsttick = np.compress(yticks >= min(ax.get_ylim()), yticks)[0]
            lasttick = np.compress(yticks <= max(ax.get_ylim()), yticks)[-1]
            ax.spines["left"].set_bounds(firsttick, lasttick)
            ax.spines["right"].set_bounds(firsttick, lasttick)
            newticks = yticks.compress(yticks <= lasttick)
            newticks = newticks.compress(newticks >= firsttick)
            ax.set_yticks(newticks)



from __future__ import annotations

from datetime import datetime
from datetime import timedelta
from types import NoneType  # type: ignore
from typing import get_type_hints
from typing import Iterator

import numpy as np
import pandas as pd


def dftoclass(name: str, df: pd.DataFrame, kind: str = "dataclass") -> None:
    ret = []
    inp = "import numpy as np"
    idt = "from datetime import datetime"
    itd = "from datetime import timedelta"
    iany = "from typing import Any"
    if kind == "dict":
        imports = {"from typing import TypedDict"}
    elif kind == "tuple":
        imports = {"from typing import NamedTuple"}
    else:
        imports = {"from dataclasses import dataclass"}
    for cname, dtype in df.dtypes.to_dict().items():
        k = dtype.kind
        if k == "f":
            ret.append((cname, "float"))
        elif k in {"i", "u"}:
            ret.append((cname, "int"))
        elif k == "c":
            ret.append((cname, "complex"))
        elif k == "b":
            ret.append((cname, "bool"))
        elif k == "S":
            ret.append((cname, "str"))
        elif k == "M":
            ret.append((cname, "datetime"))
            imports.add(idt)
        elif k == "m":
            ret.append((cname, "timedelta"))
            imports.add(itd)
        elif k == "O":
            if len(df) == 0:
                raise RuntimeError(
                    "can't determine type of object in zero row dataframe",
                )
            f = df[cname].iloc[0]
            tf = type(f)
            if tf == str:
                ret.append((cname, "str"))
            elif tf == np.ndarray:
                if len(f) > 0:
                    t = type(f[0])
                    itemt = f"  # item: {t.__name__}"
                else:
                    itemt = ""
                ret.append((cname, f"np.ndarray{itemt}"))
                imports.add(inp)
            else:
                ret.append((cname, "Any"))
                imports.add(iany)
        else:
            ret.append((cname, "Any"))
            imports.add(iany)

    attr = "\n".join(f"    {n}: {t}" for n, t in ret)
    comment = "\n# generated by dftoclass\n\n"
    KINDS = {"dict": "TypedDict", "tuple": "NamedTuple"}
    if kind in KINDS:
        typ = KINDS[kind]
        cls = [
            *imports,
            comment,
            f"class {name}({typ}):",
            attr,
        ]
    else:
        cls = [
            *imports,
            comment,
            "@dataclass",
            f"class {name}:",
            attr,
        ]
    print("\n".join(cls))


KINDS = {
    "f": float,
    "i": int,
    "u": int,
    "c": complex,
    "b": bool,
    "S": str,
    "M": datetime,
    "m": timedelta,
    "O": object,
}


def check_df_columns(df: pd.DataFrame, T: type, full: bool = True) -> set[str]:
    ret = []
    ann = get_type_hints(T)
    missing = set(ann) - set(df.columns)
    if missing:
        ret.append(f"missing columns: {missing}")

    if not full:
        return set(ret)

    n = len(df)
    for cname, dtype in df.dtypes.to_dict().items():
        if cname not in ann:  # don't care about extra columns
            continue
        k = dtype.kind
        typ = ann[cname]
        if not typ == KINDS.get(k):
            if k == "O":  # can't check
                if n == 0:
                    continue
                rtyp = type(df[cname].iloc[0])
                if rtyp == typ:
                    continue
            ret.append(f"{cname}: found type {dtype}, expecting type {typ}")
    return set(ret)


def check_pepxml_columns(df: pd.DataFrame, full: bool = True) -> set[str]:
    from .pepxmltypes import PepXMLRow

    return check_df_columns(df, PepXMLRow, full=full)


def diff(df1: pd.DataFrame, df2: pd.DataFrame, eps: float = 1e-6) -> Iterator[str]:
    if not (len(df1) == len(df2) and (df1.index == df2.index).all()):
        yield "indexes differ"
    if not (
        len(df1.columns) == len(df2.columns) and (df1.columns == df2.columns).all()
    ):
        dc = list(set(df1.columns) ^ set(df2.columns))
        yield f"columns differ: {dc}"

    cols = list(set(df1.columns) & set(df2.columns))
    rsuffix = "_o"
    df = df1[cols].join(df2[cols], rsuffix=rsuffix)
    if len(df) == 0:
        yield "can't join"
        return
    yield from diff_joined(df, rsuffix=rsuffix, eps=eps)


def diff_joined(  # noqa: C901
    df1: pd.DataFrame,
    *,
    rsuffix: str,
    eps: float = 1e-6,
) -> Iterator[str]:
    def dictok(d1, d2) -> bool:
        if not isinstance(d1, dict) and not isinstance(d2, dict):
            ret = d1 == d2
            if isinstance(ret, bool):
                return ret
            return ret.all()

        s = set(d1.keys()) == set(d2.keys())
        if not s:
            return False
        for k in d1:
            ok = dictok(d1[k], d2[k])
            if not ok:
                return False
        return True

    def ndarray_failed(arr: pd.DataFrame, col: str) -> bool:
        a1 = arr[col]
        a2 = arr[col + rsuffix]
        if a1 is None:
            return a2 is not None

        if a2 is None:
            return a1 is not None

        if a1.shape != a2.shape:
            return True
        if a1.dtype != a2.dtype:
            return True

        if len(a1) == 0:
            return False
        if a1.dtype.kind == "f":
            na = np.isnan(a1)
            if (na != np.isnan(a2)).sum() != 0:
                return True
            d = a1[~na] != a2[~na]
        else:
            d = a1 != a2  # ndarray of say str

        if isinstance(d, bool):
            for a, b in zip(a1, a2):
                ok = dictok(a, b)
                if not ok:
                    return True
            return False
        failed = d.sum() != 0

        return failed

    def check_array(s1: pd.Series, s2: pd.Series) -> int:
        t = s1.dtype
        if t != s2.dtype and t.kind != s2.dtype.kind:
            return len(s1)
        na = s1.isna()
        d = na != s2.isna()
        failed = d.sum()
        if not failed:
            if t.kind in {"f", "c", "m", "M"}:
                d = (s1[~na] - s2[~na]).abs() > eps
            else:
                d = s1[~na] != s1[~na]
            failed = d.sum()
        return failed

    def find_type(s: pd.Series) -> type | None:
        for i in range(len(s)):
            rtyp = type(s.iloc[i])
            if rtyp != NoneType:
                return rtyp
        return None

    cols = [c for c in df1.columns if not c.endswith(rsuffix)]
    rcols = [c[: -len(rsuffix)] for c in df1.columns if c.endswith(rsuffix)]

    acols = set(cols) & set(rcols)
    if not acols == set(cols):
        yield "columns differ"
    for col in acols:
        dc1 = df1[col]
        dc2 = df1[col + rsuffix]
        t = dc1.dtype

        if t != dc2.dtype:
            samekind = t.kind == dc2.dtype.kind
            yield f"{col}: types differ {t} != {dc2.dtype} {'continue checking' if samekind else 'skipping'}"
            if not samekind:
                continue

        if t.kind in {"b", "i", "u", "S", "U"}:
            failed = (dc1 != dc2).sum()

        elif t.kind in {"f", "c", "m", "M"}:
            failed = check_array(dc1, dc2)

        elif t.kind == "O":
            rtyp = find_type(dc1)

            if rtyp == np.ndarray:
                failed = (
                    pd.concat([dc1, dc2], axis=1)
                    .apply(
                        ndarray_failed,
                        args=(col,),
                        axis=1,
                    )
                    .sum()
                )

            elif rtyp == str:
                failed = check_array(dc1, dc2)

            else:
                yield f"unknown type: {col}[{rtyp or t}]"
                continue
        else:
            # kind == 'V' void
            yield f"unknown type: {col}[{t}]"
            continue

        if failed:
            yield f"{col}[{t}]: {failed}"



from __future__ import annotations

from pathlib import Path
from typing import Callable
from typing import TYPE_CHECKING

import click

from .cli import cli
from .cli import Config
from .cli import IsFile
from .cli import pass_config
from .settings import setting_options

if TYPE_CHECKING:
    from .utils import PeptideSettings


@cli.command(name="mzi-create", hidden=False)
@click.option("--force", is_flag=True, help="force creation")
@click.option(
    "-w",
    "--workers",
    type=int,
    help="number of background workers [default: half number of cpus]",
)
@click.option(
    "-o",
    "--out",
    help="directory to write [default is directory of mzML file]",
    type=click.Path(dir_okay=True, file_okay=False),
)
@click.argument("mzmlfiles", nargs=-1, type=IsFile)
def mzml_create_cmd(
    mzmlfiles: tuple[str, ...],
    out: str | None,
    workers: int | None,
    force: bool = False,
) -> None:
    """Create mz/intensity memory mapped files for mzML files"""

    from functools import partial
    from os import cpu_count

    from .utils import MzMLResourceFileLocal
    from .parallel_utils import parallel_result
    from .pymz import mzml_create
    from .logger import logger

    if workers is None:
        workers = max((cpu_count() or 1) // 2, 1)

    targets = [MzMLResourceFileLocal(m, out) for m in set(mzmlfiles)]
    todo = [t for t in targets if not t.cache_mzml_ok()] if not force else targets
    skipped = set(targets) - set(todo)
    if skipped:
        for t in skipped:
            logger.info("skipping creation of %s cache", t.original.name)
    if not todo:
        return
    if out is not None:
        pout = Path(out)
        if not pout.exists():
            pout.mkdir(parents=True, exist_ok=True)

    exe: list[Callable[[], int]] = []
    for idx, target in enumerate(todo, start=1):
        exe.append(partial(mzml_create, target, idx, len(todo)))

    ntotal = len(exe)

    for idx, _ in enumerate(parallel_result(exe, workers=workers), start=1):
        logger.info("mzml prepare task done: [%d/%d]", idx, ntotal)


@cli.command(name="run")
@click.option(
    "-w",
    "--workers",
    type=int,
    help="number of parallel workers [default: half the number of cpus]",
)
@click.option(
    "--nspectra",
    type=int,
    help="split tasks into this many spectra (see NSPECTRA configuration variable)",
)
@click.option(
    "--job-dir",
    type=click.Path(file_okay=False),
    help="directory to run job [default: directory of jobfile]",
)
@click.option(
    "--cache-dir",
    type=click.Path(file_okay=False),
    help="directory to store cache files [default: cache_dir specified in jobfile]",
)
@click.option("--force", is_flag=True, help="force (re)creation of cache files")
@click.option(
    "--no-email",
    is_flag=True,
    help="don't send any emails when finished",
)  # see runner.py
@click.option(
    "--interrupt-as-error",
    default=0,
    help="accept keyboard interrupt as error. Value will be exit code",
    hidden=True,  # only useful to for background process: see runner.py
)  # see runner.py
@click.argument("job")
@pass_config
def turnover_job_run_cmd(
    cfg: Config,
    job: str,
    force: bool = False,
    workers: int | None = None,
    nspectra: int | None = None,
    job_dir: str | None = None,
    interrupt_as_error: int = 0,
    no_email: bool = False,
    cache_dir: str | None = None,
) -> None:
    """Run a turnover job"""
    from dataclasses import replace
    from os.path import dirname
    from os import cpu_count
    from .pymz import turnover_run, turnover_prepare
    from .utils import ResultsResourceFile

    from .logger import init_logger, logger
    from .jobs import TurnoverJob
    from .config import NSPECTRA, INSPECT_URL, MAIL_TEXT

    if nspectra is None:
        nspectra = NSPECTRA

    if workers is None:
        workers = max((cpu_count() or 1) // 2, 1)

    def maybeemail(template: str | None, url: str | None = None) -> None:
        if not no_email and jobby.email and template:
            from .mailer import sendmail

            if url:
                try:
                    url = url.format(jobid=jobby.jobid)
                    url = f' <a href="{url}">See results</a>'
                except KeyError:
                    url = ""
            else:
                url = ""
            msg = template.format(job=jobby, url=url)

            try:
                sendmail(msg, jobby.email, mimetype="html")
            except OSError as e:
                logger.warning("failed to send email to %s: reason %s", jobby.email, e)

    jobby = TurnoverJob.restore(job)

    if job_dir is None:
        job_dir = dirname(job)

    if cache_dir is not None:
        jobby = replace(jobby, cache_dir=cache_dir)

    pjobdir = Path(job_dir)
    pjobdir.mkdir(exist_ok=True, parents=True)

    # send fitEnvelope warnings to a logfile if running in background
    logfile: str | Path | None
    if cfg.logfile is None and interrupt_as_error > 0:
        # this is a background job and no logfile has been specified
        logfile = pjobdir / f"{jobby.jobid}.log"
    else:
        logfile = cfg.logfile
    level: str | int = "INFO" if cfg.loglevel is None else logger.getEffectiveLevel()
    # only reinit if we haven't specified logfile
    init_logger(level=level, logfile=logfile, reinit=cfg.logfile is None)

    # lock the cache directories so we don't delete anything until results done
    # we assume the job dirctory is never deleted
    rf = jobby.to_resource_files()
    results = ResultsResourceFile(
        jobby.jobid,
        job_dir,
    )
    try:
        rf.ensure_directories()
        rf.lock(jobby.jobid)
        turnover_prepare(rf, force=force, workers=workers)

        turnover_run(
            jobby,
            results,  # where to store stuff for this job
            workers=workers,
            nspectra=nspectra,
            save_subset=True,
        )
        # interrupt as_error != 0 means we are running as the
        # background job for the website.
        if interrupt_as_error:
            jobby = replace(jobby, status="finished")
            jobby.save(job_dir)  # save a copy of the job
    except Exception as e:
        logger.error("turnover %s failed. reason: %s", jobby.jobid, e)
        results.cleanup()
        maybeemail('Protein Turnover Job "{job.job_name}" has <b>failed</b>!')
        raise e
    except KeyboardInterrupt as e:
        results.cleanup()
        if interrupt_as_error:
            logger.error("turnover %s killed", jobby.jobid)
            raise SystemExit(interrupt_as_error) from e
        raise e
    finally:
        rf.unlock(jobby.jobid)

    maybeemail(MAIL_TEXT, INSPECT_URL)


@cli.command(name="job")
@click.option(
    "-o",
    "--out",
    help="filename to write file to",
    type=click.Path(dir_okay=False, file_okay=True),
)
@click.option("-j", "--jobid", help="job id")
@click.option("-n", "--name", help="job name")
@click.option(
    "-x",
    "--no-check",
    is_flag=True,
    help="don't do any sanity checks on the files",
)
@click.argument("pepxml", type=IsFile)
@click.argument("protxml", type=IsFile)
@click.argument("mzmlfiles", nargs=-1, type=IsFile)
@setting_options
def turnover_job(
    pepxml: str,
    protxml: str,
    mzmlfiles: list[str],
    no_check: bool,
    name: str | None,
    jobid: str | None,
    settings: PeptideSettings,
    out: str | None = None,
) -> None:
    """Create a turnover job file"""
    from .jobs import TurnoverJob, jobidkey
    from .pepxml import scan_spectra
    from .protxml import scan_proteins
    from .pymz import scan_mzml_spectra
    from .jobs import slugify

    if not no_check:
        missing = []
        for f in [pepxml, protxml, *mzmlfiles]:
            if not Path(f).exists():
                missing.append(f)
        if missing:
            click.secho(
                f"These are not files: {', '.join(missing)}",
                fg="red",
                err=True,
            )
            raise click.Abort()

        n = sum(1 for _ in scan_spectra(Path(pepxml)))
        if n == 0:
            raise click.BadParameter(
                f"peptide XML {pepxml} has no spectra",
                param_hint="pepxml",
            )
        mzmlfiles = list(set(mzmlfiles))
        n = sum(1 for _ in scan_proteins(Path(protxml)))
        if n == 0:
            raise click.BadParameter(
                f"protein XML {protxml} has no proteins",
                param_hint="protxml",
            )
        failed = []
        for mzml in mzmlfiles:
            n = sum(1 for _ in scan_mzml_spectra(Path(mzml)))
            if n == 0:
                failed.append(mzml)
        if failed:
            raise click.BadParameter(
                f'mzML "{ ", ".join(failed) }" has no spectra',
                param_hint="mzmlfiles",
            )

    if not jobid:
        if name:
            jobid = slugify(name)
        if not jobid or len(jobid) < 4:
            jobid = jobidkey(pepxml, protxml, mzmlfiles)

    job_name = name or jobid

    if out is None:
        filename = Path(f"{jobid}.toml")
    else:
        filename = Path(out)
    if not filename.parent.exists():
        filename.parent.mkdir(parents=True, exist_ok=True)

    job = TurnoverJob(
        job_name=job_name,
        pepxml=[pepxml],
        protxml=protxml,
        mzmlfiles=mzmlfiles,
        cache_dir=str(filename.parent),
        jobid=jobid,
        settings=settings,
    )

    jobfile = job.save(filename.parent, filename.name)
    click.secho(f'written job to "{jobfile}"', fg="yellow")


@cli.command(name="prepare", hidden=True)
@click.option(
    "-o",
    "--out",
    help="directory to write",
    type=click.Path(dir_okay=True, file_okay=False),
)
@click.option("--force", is_flag=True, help="force creation")
@click.argument("job", type=IsFile)
def turnover_job_prepare(
    job: str,
    out: str | None,
    force: bool = False,
) -> None:
    """Prepare pepXML and mzML cache files"""
    from dataclasses import replace
    from .jobs import TurnoverJob
    from .pymz import turnover_prepare

    jobby = TurnoverJob.restore(job)
    if out is not None:
        jobby = replace(jobby, cache_dir=out)

    rf = jobby.to_resource_files()
    turnover_prepare(
        rf,
        force=force,
    )


@cli.command(hidden=True)
@click.option("--verbose", is_flag=True, help="show more data")
@click.argument("filenames", nargs=-1, type=IsFile)
def hash(filenames: list[str], verbose: bool) -> None:
    "Calculate sha256 hashes from files"
    from datetime import datetime
    from .utils import hash256, human

    for name in [Path(n) for n in set(filenames)]:
        s = datetime.now()
        h = hash256(name)
        e = datetime.now()

        if verbose:
            size = name.stat().st_size
            dt = e - s
            bs = int(size / dt.total_seconds())
            click.echo(f"{h} [{dt}]: {human(size):>8}: {human(bs):>8}/s {name}")
        else:
            click.echo(f"{h} {name}")



from __future__ import annotations

from sys import byteorder

# change this to cache bust
VERSION = 1

EXT = "parquet"
PEPXML = f".pep-v{VERSION}.{EXT}"
PROTXML = f".prot-v{VERSION}.{EXT}"
MZMAP = f".{byteorder}-v{VERSION}.mzi"
DINOSAUR = f".features-v{VERSION}.{EXT}"
MZML = f".mzml-v{VERSION}.{EXT}"
EICS = f".eics-v{VERSION}.{EXT}"
RESULT_EXT = ".sqlite"  # {EXT}"  # f".v{VERSION}.{EXT}"



from __future__ import annotations

import hashlib
import math
from dataclasses import asdict
from dataclasses import dataclass
from functools import reduce
from pathlib import Path
from typing import Any
from typing import Literal
from typing import TypedDict

import numpy as np
import pandas as pd
from scipy.fft import fft
from scipy.fft import ifft

from .exts import DINOSAUR
from .exts import EICS
from .exts import EXT
from .exts import MZMAP
from .exts import MZML
from .exts import PEPXML
from .exts import PROTXML
from .exts import RESULT_EXT
from .types.pepxmltypes import PepXMLRow
from .types.pepxmltypes import PepXMLRunRow


@dataclass
class PeptideSettings:
    rtTolerance: float = 15.0  # seconds
    mzTolerance: float = 1e-5  # ppm
    labelledIsotopeNumber: int = 15
    labelledElement: str = "N"
    maximumLabelEnrichment: float = 0.95
    retentionTimeCorrection: Literal["UseInSample", "SimpleMedian"] = "SimpleMedian"
    useObservedMz: bool = False
    minProbabilityCutoff: float = 0.8
    # fdrMaximum: float = 0.0
    # psmFilters: Filter | None = None

    def hash(self) -> str:
        from hashlib import md5

        m = md5()
        for k, v in asdict(self).items():
            if isinstance(v, float):
                v = round(v, 6)
            m.update(f"{k}={v}".encode())
        return m.hexdigest()

    def __hash__(self) -> int:
        return int(self.hash(), 16)


NPA = np.array


def NPI(list_of_ints: list[int]) -> np.ndarray[Any, np.dtype[np.int32]]:
    return np.array(list_of_ints, dtype=np.int32)


NAMES = np.array(["C", "H", "O", "N", "P", "S", "Se"])
# "C", "H", "O" ,"N", "P", "S" , "Se"
# minus H20
#   "U": NPI([3,5,1,1,0,0,1])
aminoAcids = {
    "A": NPI([3, 5, 1, 1, 0, 0, 0]),
    "C": NPI([3, 5, 1, 1, 0, 1, 0]),
    "D": NPI([4, 5, 3, 1, 0, 0, 0]),
    "E": NPI([5, 7, 3, 1, 0, 0, 0]),
    "F": NPI([9, 9, 1, 1, 0, 0, 0]),
    "G": NPI([2, 3, 1, 1, 0, 0, 0]),
    "H": NPI([6, 7, 1, 3, 0, 0, 0]),
    "I": NPI([6, 11, 1, 1, 0, 0, 0]),
    "K": NPI([6, 12, 1, 2, 0, 0, 0]),
    "L": NPI([6, 11, 1, 1, 0, 0, 0]),
    "M": NPI([5, 9, 1, 1, 0, 1, 0]),
    "N": NPI([4, 6, 2, 2, 0, 0, 0]),
    "P": NPI([5, 7, 1, 1, 0, 0, 0]),
    "Q": NPI([5, 8, 2, 2, 0, 0, 0]),
    "R": NPI([6, 12, 1, 4, 0, 0, 0]),
    "S": NPI([3, 5, 2, 1, 0, 0, 0]),
    "T": NPI([4, 7, 2, 1, 0, 0, 0]),
    "V": NPI([5, 9, 1, 1, 0, 0, 0]),
    "W": NPI([11, 10, 1, 2, 0, 0, 0]),
    "Y": NPI([9, 9, 2, 1, 0, 0, 0]),
    "a": NPI([2, 2, 1, 0, 0, 0, 0]),
    "c": NPI([5, 8, 2, 2, 0, 1, 0]),
    "m": NPI([5, 9, 2, 1, 0, 1, 0]),
    "U": NPI([3, 5, 1, 1, 0, 0, 1]),
}


class AtomicProperties(TypedDict):
    isotopeNr: np.ndarray[Any, np.dtype[np.int32]]
    abundance: np.ndarray[Any, np.dtype[np.float64]]
    mass: np.ndarray[Any, np.dtype[np.float64]]


atomicProperties: dict[str, AtomicProperties] = dict(
    C=dict(
        isotopeNr=NPI([12, 13]),
        mass=NPA([12.0, 13.0033548378]),
        abundance=NPA([0.9889, 0.0111]),
    ),
    H=dict(
        isotopeNr=NPI([1, 2]),
        mass=NPA([1.0078250321, 2.0141017780]),
        abundance=NPA([0.9998, 0.0001]),
    ),
    O=dict(
        isotopeNr=NPI([16, 17, 18]),
        mass=NPA([15.9949146, 16.9991312, 17.9991603]),
        abundance=NPA([0.9976, 0.0004, 0.0020]),
    ),
    N=dict(
        isotopeNr=NPI([14, 15]),
        mass=NPA([14.0030740052, 15.0001088984]),
        abundance=NPA([0.99633, 0.00367]),
    ),
    # need entry for formula...
    P=dict(isotopeNr=NPI([31]), mass=NPA([30.97376163]), abundance=NPA([1.0])),
    # out of (isotope number) order abundance...
    S=dict(
        isotopeNr=NPI([32, 34, 33, 36]),
        mass=NPA([31.97207070, 33.96786665, 32.97145843, 35.96708062]),
        abundance=NPA([0.9502, 0.0421, 0.0075, 0.0002]),
    ),
    # Se=dict(
    #     isotopeNr=NPI([74, 76, 77, 78, 80, 82]),
    #     mass=NPA(
    #         [73.9224764, 75.9192136, 76.9199140, 77.9173091, 79.9165213, 81.9166994],
    #     ),
    #     abundance=NPA([0.0089, 0.0937, 0.0763, 0.2377, 0.4961, 0.0873]),
    # ),
    Se=dict(
        isotopeNr=NPI([80, 78, 76, 82, 77, 74]),
        mass=NPA(
            [79.9165213, 77.9173091, 75.9192136, 81.9166994, 76.919914, 73.9224764],
        ),
        abundance=NPA([0.4961, 0.2377, 0.0937, 0.0873, 0.0763, 0.0089]),
    ),
)


def getNr(element: str) -> int:
    """get isotope number for most abundant element"""
    if element not in atomicProperties:
        raise ValueError(f'unknown element: "{element}"')
    at = atomicProperties[element]
    a = at["abundance"]
    return int(at["isotopeNr"][a.argmax()])


def okNr(element: str) -> set[int]:
    at = atomicProperties[element]
    a = at["isotopeNr"]
    return {int(i) for i in a}


def roundit(n: float, nsig: int = 3) -> int:
    from math import log10

    assert n > 0, str(n)
    pwr = round(log10(n))
    pwr = max(0, pwr - nsig)
    num = 10**pwr
    return round(n / num) * num


# return C H O N P S Se count list
def peptideFormula(peptideSequence: str) -> np.ndarray[Any, np.dtype[np.int32]]:
    startComp = NPI([0, 2, 1, 0, 0, 0, 0])
    return reduce(
        lambda total, aa: total + aminoAcids[aa],
        list(peptideSequence),
        startComp,
    )


def get_abundance(setting: PeptideSettings) -> float:
    a = atomicProperties[setting.labelledElement]
    b = a["isotopeNr"] == setting.labelledIsotopeNumber
    return a["abundance"][b][0]


def isotopicDistribution(
    peptideSequence: str,
    abundances: dict[str, np.ndarray] | None = None,
    abundanceCutoff: float = 1e-10,
    maxMass: int = 100,
) -> np.ndarray[Any, np.dtype[np.float64]]:
    """return abundance of "heavy" peptides"""
    if abundances is None:
        abundances = {}

    formula = peptideFormula(peptideSequence)

    maxElements = np.sum(formula > 0, dtype=int)

    A = np.zeros((maxElements, maxMass))

    elements = []
    for i, e in enumerate(NAMES):
        n = formula[i]
        if n > 0:
            elements.append(n)
            if e in abundances:
                a = abundances[e]
            else:
                # use natural abundance
                a = atomicProperties[e]["abundance"]
            A[len(elements) - 1, 0 : len(a)] = a

    tA = fft(A)
    ptA = np.ones((maxMass,), dtype=np.complex128)

    for i in range(maxElements):
        ptA *= tA[i] ** elements[i]

    riptA = np.real(ifft(ptA))
    mx = np.max(np.where(riptA > abundanceCutoff))
    riptA = riptA[0 : int(mx) + 1]
    return np.fmax(riptA, 0.0)
    # return np.where(riptA > 0.0, riptA, 0.0)


def labelledAtomicAbundances(
    labelledIsotopeNumber: int,
    element: str,
    abundance: float,
) -> dict[str, np.ndarray]:
    if element in atomicProperties:
        ap = atomicProperties[element]
        if labelledIsotopeNumber in ap["isotopeNr"]:
            a = ap["abundance"]
            label = ap["isotopeNr"] == labelledIsotopeNumber
            adjustedAbundance: np.ndarray = (1 - abundance) * a / np.sum(a[~label])

            adjustedAbundance[label] = abundance
            return {element: adjustedAbundance}
    raise ValueError(f"unknown element {element}[{labelledIsotopeNumber}]")


def labelledAtomicAbundances2(settings: PeptideSettings) -> dict[str, np.ndarray]:
    return labelledAtomicAbundances(
        settings.labelledIsotopeNumber,
        settings.labelledElement,
        settings.maximumLabelEnrichment,
    )


def theoretical_dist(
    peptide: str,
    labelEnrichment: float,
    settings: PeptideSettings,
) -> np.ndarray:
    from .config import ABUNDANCE_CUTOFF  # pylint: disable=import-outside-toplevel

    abundances = labelledAtomicAbundances(
        settings.labelledIsotopeNumber,
        settings.labelledElement,
        labelEnrichment,
    )
    theoreticalDist = isotopicDistribution(
        peptide,
        abundances,
        abundanceCutoff=ABUNDANCE_CUTOFF,
    )
    return theoreticalDist


def okiso(settings: PeptideSettings) -> bool:
    props = atomicProperties[settings.labelledElement]
    isotopeNr, abundance = props["isotopeNr"], props["abundance"]
    idx = (isotopeNr == settings.labelledIsotopeNumber)[0]
    ab = (abundance == np.max(abundance))[0]
    return idx != ab


def isoMzDiff(pep: PepXMLRow, settings: PeptideSettings) -> float:
    props = atomicProperties[settings.labelledElement]
    mass, abundance = props["mass"], props["abundance"]
    isoDiff = (
        mass[props["isotopeNr"] == settings.labelledIsotopeNumber]
        - mass[abundance == np.max(abundance)]
    )
    isoMzDiff = isoDiff[0] / pep.assumed_charge
    return isoMzDiff


def eic_mzranges(pep: PepXMLRunRow, settings: PeptideSettings) -> np.ndarray:
    # requires
    # maxIso, mz, assumed_charge and
    # (settings.{labelledElement,labelledIsotopeNumber}, mzTolerance)
    isoDiff = isoMzDiff(pep, settings)
    mz = pep.observed_mz if settings.useObservedMz else pep.mz
    tol = mz * settings.mzTolerance
    isoRange = range(-1, round(pep.maxIso) + 1)

    # NOTE: pep.maxIso might be a floating point number if
    # pep is actually a pd.Series
    mzranges = [
        [
            mz + iso * isoDiff - tol,
            mz + iso * isoDiff + tol,
        ]
        for iso in isoRange
    ]
    return np.array(mzranges, dtype=np.float32)


def resize(a: np.ndarray, width: int) -> np.ndarray:
    d = width - len(a)
    if d > 0:
        return np.pad(a, (0, d), "constant", constant_values=(0, 0))
    if d < 0:
        return a[:width]
    return a


def ensure_pos(a: np.ndarray) -> np.ndarray:
    # use np.fmax if we want to set a[i] to zero if a[i] is NaN
    return np.maximum(a, 0.0)
    # return np.where(a >= 0, a, 0.0)


def human(num: int, suffix: str = "B", scale: int = 1) -> str:
    if not num:
        return f"0{suffix}"
    num *= scale
    magnitude = int(math.floor(math.log(abs(num), 1000)))
    val = num / math.pow(1000, magnitude)
    if magnitude > 7:
        return "{:.1f}{}{}".format(val, "Y", suffix)
    return "{:3.1f}{}{}".format(
        val,
        ["", "k", "M", "G", "T", "P", "E", "Z"][magnitude],
        suffix,
    )


def hash256(filename: Path, bufsize: int = 4096 * 5) -> str:
    sha256_hash = hashlib.sha256()
    with filename.open("rb") as f:
        # Read and update hash string value in blocks of 4K
        for byte_block in iter(lambda: f.read(bufsize), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


def fileok(cached: list[Path], original: Path) -> bool:
    if not cached:
        return False
    if not original.is_file():
        return False
    omtime = original.stat().st_mtime
    for cache in cached:
        if not cache.is_file() or omtime > cache.stat().st_mtime:
            return False
    return True


class BaseResourceFile:
    def __init__(
        self,
        original: Path | str,
        cache_dir: Path | str | None = None,
    ):
        self.original = Path(original).expanduser().resolve()
        if cache_dir is None:
            self.cache_dir = self.original.parent
        else:
            self.cache_dir = Path(cache_dir).expanduser().resolve()
        self._hash: str | None = None

    @property
    def name(self) -> str:
        return self.original.name

    @property
    def exists(self) -> bool:
        return self.original.exists()

    def extok(self, *ext: str) -> bool:
        targets = [self.cache_file(e) for e in ext]
        return fileok(targets, self.original)

    def cache_file(self, ext: str) -> Path:
        # return self.cache_dir.joinpath(self.name + ext)
        return self.cache_dir.joinpath(self.hash + ext)

    def touch(self) -> bool:
        return False

    def cache_ok(self) -> bool:
        return False

    @property
    def hash(self) -> str:
        if self._hash is not None:
            return self._hash

        self._hash = self.get_hash()
        return self._hash

    def get_hash(self) -> str:
        return hash256(self.original)

    def cleanup(self):
        for pth in self.all_cache_files():
            try:
                pth.unlink(missing_ok=True)
            except OSError:
                pass

    def all_cache_files(self) -> list[Path]:
        return []


class MzMLResourceFile(BaseResourceFile):
    @property
    def runName(self) -> str:
        """Name possibly in pep.xml file as <spectrum_query spectrum="{runName}.{start_scan}.{end_scan}.0 ...>"""
        return self.original.stem

    def cache_mzml(self) -> Path:
        return self.cache_file(MZML)

    def cache_memmap(self) -> Path:
        return self.cache_file(MZMAP)

    def cache_dinosaur(self) -> Path:
        return self.cache_file(DINOSAUR)

    def cache_mzml_ok(self) -> bool:
        from .dinosaur.dinosaur import DinoRunner

        ok = self.extok(MZMAP, MZML)
        if not ok:
            return False
        ok = self.extok(DINOSAUR)
        if ok:
            return True

        dino = DinoRunner.from_config()
        # no dino file check if we can generate it...
        if dino is not None and dino.can_run():
            return False
        # ok we can't generate it so cache is "fine..."
        return True

    def cache_ok(self) -> bool:
        return self.cache_mzml_ok()

    def touch(self) -> bool:
        if self.cache_ok():
            self.cache_memmap().touch()
            self.cache_mzml().touch()
            return True
        return False

    def all_cache_files(self) -> list[Path]:
        return [self.cache_mzml(), self.cache_memmap()]


class MzMLResourceFileLocal(MzMLResourceFile):
    def __init__(
        self,
        original: Path | str,
        directory: Path | str | None = None,
        name: str | None = None,
    ):
        super().__init__(original, directory)
        if name is None:
            name = self.runName
        self.outname = name

    def cache_file(self, ext: str) -> Path:
        # return self.cache_dir.joinpath(self.name + ext)
        return self.cache_dir.joinpath(self.outname + ext)

    def cache_mzml(self) -> Path:
        return self.cache_file("." + EXT)

    def cache_memmap(self) -> Path:
        return self.cache_file(".mzi")

    def cache_mzml_ok(self) -> bool:
        return self.extok("." + EXT, ".mzi")


class PepXMLResourceFile(BaseResourceFile):
    def cache_pepxml(self) -> Path:
        return self.cache_file(PEPXML)

    def cache_pepxml_ok(self) -> bool:
        return self.extok(PEPXML)

    def cache_ok(self) -> bool:
        return self.cache_pepxml_ok()

    def touch(self) -> bool:
        if self.cache_ok():
            self.cache_pepxml().touch()
            return True
        return False

    def all_cache_files(self) -> list[Path]:
        return [self.cache_pepxml()]


class ProtXMLResourceFile(BaseResourceFile):
    def cache_protxml(self) -> Path:
        return self.cache_file(PROTXML)

    def cache_protxml_ok(self) -> bool:
        return self.extok(PROTXML)

    def cache_ok(self) -> bool:
        return self.cache_protxml_ok()

    def touch(self) -> bool:
        if self.cache_ok():
            self.cache_protxml().touch()
            return True
        return False

    def all_cache_files(self) -> list[Path]:
        return [self.cache_protxml()]


class ResultsResourceFile(BaseResourceFile):
    def cache_result(self) -> Path:
        return self.cache_file(RESULT_EXT)

    def cache_pepxml(self) -> Path:
        return self.cache_file(PEPXML)

    def cache_envelope(self) -> Path:
        return self.cache_file(EICS)

    def cache_file(self, ext: str) -> Path:
        return self.cache_dir.joinpath(self.original.name + ext)

    def all_cache_files(self) -> list[Path]:
        return [self.cache_result(), self.cache_pepxml(), self.cache_envelope()]

    def has_result(self) -> bool:
        return self.cache_result().exists()


class ResourceFiles:
    def __init__(
        self,
        pepxmls: list[PepXMLResourceFile],
        protxml: ProtXMLResourceFile,
        mzmlfiles: list[MzMLResourceFile],
    ):
        self.pepxmls = pepxmls
        self.protxml = protxml
        self.mzmlfiles = mzmlfiles
        self.cache_dirs = {
            *[s.cache_dir for s in pepxmls],
            *[s.cache_dir for s in mzmlfiles],
        }

    def lock(self, job_id: str) -> None:
        for c in self.cache_dirs:
            c.joinpath(job_id + ".lock").touch()

    def unlock(self, job_id: str) -> None:
        for c in self.cache_dirs:
            c.joinpath(job_id + ".lock").unlink(missing_ok=True)  # type: ignore

    def is_locked(self) -> bool:
        return len([f for c in self.cache_dirs for f in c.glob("*.lock")]) > 0

    def todo(self) -> list[BaseResourceFile]:
        ret: list[BaseResourceFile] = []
        for pepxml in self.pepxmls:
            if not pepxml.cache_pepxml_ok():
                ret.append(pepxml)
        for mzml in self.mzmlfiles:
            if not mzml.cache_mzml_ok():
                ret.append(mzml)
        return ret

    def ensure_directories(self) -> None:
        for cd in self.cache_dirs:
            if not cd.is_dir():
                cd.mkdir(parents=True, exist_ok=True)


def rmfiles(files: list[Path]) -> None:
    from contextlib import suppress

    for f in files:
        with suppress(OSError):
            Path(f).unlink()


def rehydrate_turnover(turn: pd.DataFrame) -> pd.DataFrame:
    if "modifications" in turn.columns and "modcol" not in turn.columns:
        turn["modcol"] = turn["modifications"].apply(Apply.modcol)
    if "eics" in turn.columns and "eics_shape" in turn.columns:
        turn["eics"] = turn[["eics", "eics_shape"]].apply(
            Apply.eics_reshape,
            axis=1,
        )
        turn.drop(columns=["eics_shape"], inplace=True)
    if "mzranges" in turn.columns:
        turn["mzranges"] = turn["mzranges"].apply(lambda mz: mz.reshape((-1, 2)))
    if "isotopeRegressions" in turn.columns:
        turn["isotopeRegressions"] = turn["isotopeRegressions"].apply(
            lambda mz: mz.reshape((-1, 2)),
        )
    return turn


# def read_rds(filename: str) -> pd.DataFrame:
#     df = IO(filename).read_df()
#     df = rehydrate_turnover(df)
#     return df


def duplicate_rows(df: pd.DataFrame, on: list[str]) -> pd.DataFrame:
    return df[df.duplicated(on, keep=False)]


def fdr(df: pd.DataFrame) -> float:
    decoy = df["is_decoy"].sum()
    denom = len(df) - decoy
    return float(decoy) / denom if denom > 0 else np.inf


def calculate_fdrs(
    df: pd.DataFrame,
    nbins: int = 50,
    score: str = "peptideprophet_probability",
) -> pd.Series:
    """Requires peptideprophet_probability, is_decoy"""
    # df['fdr'] = calculate_fdrs(df)
    from math import ceil

    decoys = df["is_decoy"]
    pbins = df[score].apply(lambda p: ceil(p * nbins))

    mx = int(pbins.max())

    fdrs: dict[int, float] = {}
    for binno in range(mx + 1):
        target = pbins >= binno
        n = target.sum()
        if n == 0:
            fdrs[binno] = fdrs.get(binno - 1, 0.0)
            continue
        ndecoys = decoys[target].sum()
        denom = n - ndecoys
        fdrs[binno] = ndecoys / denom if denom > 0 else np.inf
    return pbins.apply(lambda binno: fdrs[binno]).astype(np.float32)


class Apply:
    @staticmethod
    def eics_reshape(s: pd.Series) -> np.ndarray:
        return s["eics"].reshape(s["eics_shape"])

    @staticmethod
    def eics_reshape_np(s: pd.Series) -> np.ndarray:
        return np.array(s["eics"]).reshape(s["eics_shape"])

    @staticmethod
    def modcol(mods: list[dict], n: int = 3) -> str:
        mods = sorted(mods, key=lambda d: d["position"])
        return ":".join([f"{d['mass']:.{n}f}@{d['position']}" for d in mods])

    @staticmethod
    def binrt(rt: float, rttol: float = 15.0) -> int:
        return round(rt / rttol)

    @staticmethod
    def binmz(mz: float, mztol: float = 10e-6) -> int:
        return round(mz / mztol)


class IO:
    def __init__(self, filename: Path | str, df: pd.DataFrame | None = None):
        self.df = df
        self.filename = Path(filename)

    def save_df(self, index: bool = False) -> None:
        assert self.df is not None
        try:
            self.df.to_parquet(self.filename, index=index)

        except (Exception, KeyboardInterrupt):
            rmfiles([self.filename])  # cleanup
            raise

    def read_df(self, columns: list[str] | None = None) -> pd.DataFrame:
        if self.filename.name.endswith(".feather"):
            return pd.read_feather(self.filename, columns=columns, use_threads=True)
        return pd.read_parquet(self.filename, columns=columns, use_threads=True)


def df_can_write(fmt: str) -> bool:
    from io import BytesIO

    df = pd.DataFrame({"x", [0]})  # pylint: disable=unhashable-member
    func = getattr(df, f"to_{fmt}", None)
    if func is None:
        return False
    try:
        func(BytesIO())
        return True
    except ModuleNotFoundError:
        return False


def df_can_write_parquet() -> bool:
    return df_can_write("parquet")


def df_can_write_excel() -> bool:
    return df_can_write("excel")


def getsize(fname: Path) -> int:
    return fname.stat().st_size



from __future__ import annotations

import math
from typing import Any

import numpy as np
import pandas as pd
from sqlalchemy import Column
from sqlalchemy import func
from sqlalchemy import inspect
from sqlalchemy import select
from sqlalchemy import Table
from sqlalchemy import text
from sqlalchemy.engine import Engine


def dataframe_sql(df: pd.DataFrame) -> str:
    res = []
    for cname, dtype in df.dtypes.to_dict().items():
        k = dtype.kind
        if k == "f":
            a = "Float"
        elif k == "i":
            a = "Integer"
        elif k == "b":
            a = "Boolean"
        else:
            typ = type(df[cname].iloc[0])
            if typ == str:
                m = df[cname].apply(len).max()
                if m > 4096:
                    a = "Text"
                else:
                    a = f"String({m})"
            else:
                a = "JSON"
        r = f"    {cname} = Column({a})"
        res.append(r)
    return "\n".join(res)


def check_missing_columns(df: pd.DataFrame, Orm: type) -> set[str]:
    i: Table = inspect(Orm)
    dtypes = {c.name for c in i.columns}
    pepcols = {str(s) for s in df.columns}
    return dtypes - pepcols


def dehydrate_floatarray(arr: np.ndarray) -> list[float]:
    return [float(f) for f in arr]


def dehydrate_intarray(arr: np.ndarray) -> list[int]:
    return [int(f) for f in arr]


def rehydrate_pep(df: pd.DataFrame) -> pd.DataFrame:
    df["modifications"] = df.modifications.apply(
        lambda strarr: np.array(strarr, dtype=object),
    )
    return df


def dehydrate_peptide_col(peptide_list: list[dict[str, Any]]) -> list[dict[str, Any]]:
    def fix(v: dict[str, Any]) -> dict[str, Any]:
        if isinstance(v, np.ndarray):
            return dehydrate_peptide_col(v)
        if isinstance(v, dict):
            return {k: fix(vv) for k, vv in v.items()}
        return v

    return [fix(value) for value in peptide_list]


def dehydrate_prot(df: pd.DataFrame) -> pd.DataFrame:
    df["peptide"] = df.peptide.apply(dehydrate_peptide_col)
    df["indistinguishable_protein"] = df.indistinguishable_protein.apply(list)
    df["unique_stripped_peptides"] = df.unique_stripped_peptides.apply(list)
    return df


def secondary(
    pep: pd.DataFrame,
    prot: pd.DataFrame,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    exploded = prot[["proteinid", "protein_name", "unique_stripped_peptides"]].explode(
        "unique_stripped_peptides",
    )
    dfm = pep[["peptideid", "peptide"]].merge(
        exploded,
        how="inner",
        left_on="peptide",
        right_on="unique_stripped_peptides",
    )
    sec = dfm[["peptideid", "proteinid"]]
    names = (
        dfm[["peptideid", "protein_name"]]
        .groupby("peptideid")
        .agg({"protein_name": lambda s: sorted(set(s))})
    )
    names = names.reset_index().rename(columns=dict(protein_name="protein_names"))
    return sec, names


def dehydrate_peptides(df: pd.DataFrame) -> pd.DataFrame:
    if "result_index" in df.columns:
        df.rename(columns=dict(result_index="peptideid"), inplace=True)
    for col in ["proteins", "protein_descr", "modifications"]:
        if col in df.columns:
            df[col] = df[col].apply(list)
    for col in [
        "mzranges",
        "eics",
        # "isotopeRegressions",
        "adjustedRsq",
        "isotopeEnvelopes",
        "monoFitParams",
        "heavyDistribution",
        "labelledEnvelopes",
        "heavyFitParams",
        "theoreticalDist",
    ]:
        if col in df.columns:
            df[col] = df[col].apply(dehydrate_floatarray)
    for col in ["eics_shape"]:
        if col in df.columns:
            df[col] = df[col].apply(dehydrate_intarray)
    return df


def rehydrate_peptides(turn: pd.DataFrame) -> pd.DataFrame:
    from ..utils import Apply

    if "eics" in turn.columns and "eics_shape" in turn.columns:
        turn["eics"] = turn[["eics", "eics_shape"]].apply(
            Apply.eics_reshape_np,
            axis=1,
        )
        turn.drop(columns=["eics_shape"], inplace=True)
    if "mzranges" in turn.columns:
        turn["mzranges"] = turn["mzranges"].apply(
            lambda mz: np.array(mz).reshape((-1, 2)),
        )
    if "isotopeRegressions" in turn.columns:
        turn["isotopeRegressions"] = turn["isotopeRegressions"].apply(
            lambda iso: np.array(iso).reshape((-1, 2)),
        )
    return turn


def show_table_create(name: str, engine: Engine) -> str:
    with engine.connect() as conn:
        return str(
            conn.execute(
                text(f"SELECT sql FROM sqlite_schema WHERE name = '{name}'"),
            ).scalar(),
        )


def round10(x: float | int) -> float | int:
    if x == 0 or not math.isfinite(x):
        return x
    digits = math.ceil(math.log10(abs(x)))
    sign = 1 if x > 0 else -1
    return sign * 10**digits


def std(engine: Engine, col: Column) -> float:
    q = select(
        func.avg(col * col),
        func.avg(col),
        func.count(col),  # pylint: disable=not-callable
    )
    # q = q.select_from(col.table)
    with engine.connect() as conn:
        xx, x, c = conn.execute(q).one()
        r = c / (c - 1) if c > 1 else 1
        v = (xx - x * x) * r
        return math.sqrt(v)



from __future__ import annotations

from dataclasses import replace
from functools import wraps
from typing import Any
from typing import Callable

import click


# from https://stackoverflow.com/questions/56185880/commands-with-multiple-common-options-going-into-one-argument-using-custom-decor
# This drags in scipy and pandas and numpy from imports above
# *and* depends on atomicProperties via getNr which drags in numpy again


def setting_options(func: Callable[..., Any]) -> Callable[..., Any]:
    # don't want to import .utils since this will drag in scipy, numpy etc
    atoms = ["C", "H", "O", "N", "P", "S"]

    @wraps(func)
    @click.option(
        "-e",
        "--element",
        "labelledElement",
        type=click.Choice(atoms),
        help="heavy isotope",
    )
    @click.option(
        "--nr",
        "labelledIsotopeNumber",
        type=click.IntRange(1),
        help="isotope number",
    )
    @click.option(
        "--mztol",
        "mzTolerance",
        type=click.FloatRange(0.0),
        help="m/z tolerance in ppm",
    )
    @click.option(
        "--rttol",
        "rtTolerance",
        type=click.FloatRange(0.0),
        help="retention time tolerance in seconds",
    )
    def distill_settings(
        labelledElement: str,
        labelledIsotopeNumber: int,
        mzTolerance: float,
        rtTolerance: float,
        **kwargs: Any,
    ) -> Any:
        # this function is being called so we can import now
        from .utils import PeptideSettings, getNr, okNr

        settings = PeptideSettings()

        def rep(**kwargs: Any) -> PeptideSettings:
            return replace(settings, **kwargs)

        if labelledElement is not None:
            settings = rep(
                labelledElement=labelledElement,
                labelledIsotopeNumber=getNr(labelledElement),
            )
        if labelledIsotopeNumber is not None:
            if labelledIsotopeNumber not in okNr(settings.labelledElement):
                raise click.BadParameter(
                    f"unknown isotope {settings.labelledElement}[{labelledIsotopeNumber}]",
                    param_hint="labelledIsotopeNumber",
                )

            settings = rep(
                labelledIsotopeNumber=labelledIsotopeNumber,
            )
        if mzTolerance is not None:
            settings = rep(mzTolerance=mzTolerance / 1.0e6)

        if rtTolerance is not None:
            settings = rep(rtTolerance=rtTolerance)

        kwargs["settings"] = settings
        return func(**kwargs)

    return distill_settings



from __future__ import annotations

from itertools import islice
from pathlib import Path
from typing import Any
from typing import Iterator

import pandas as pd
from more_itertools import ichunked
from pymzml.run import Reader
from pyteomics.pepxml import PepXML

from .logger import log_iterator
from .pepxml import chop_spectrum

# https://github.com/levitsky/pyteomics/blob/b9fabac991d2da524b4700af871e56f6a255f924/pyteomics/pepxml.py#L349
# taken from pyteomics.pepxml:DataFrame


def fix_mods(mods: list[dict]) -> Iterator[dict]:
    for m in mods:
        if "variable" in m:
            yield dict(
                mass=m["mass"],
                position=m["position"],
                massdiff=m["variable"],
            )

        elif "static" in m:
            yield dict(
                mass=m["mass"],
                position=m["position"],
                massdiff=m["static"],
            )
        else:
            # both these are optional but one must exist?
            yield dict(mass=m["mass"], position=m["position"])


def dict_to_series(item: dict[str, Any]) -> dict[str, Any]:
    info: dict[str, Any] = {}
    for k, v in item.items():
        if isinstance(v, (str, int, float)):
            info[k] = v
    if "search_hit" in item:
        sh = item["search_hit"][0]  # first element
        proteins = sh.pop("proteins")

        # list of dicts to dict of lists
        prot_dict: dict[str, list[Any]] = {}
        for p in proteins:
            for k in p:
                prot_dict[k] = []
        for p in proteins:
            for k, v in prot_dict.items():
                v.append(p.get(k))
        prot_dict["proteins"] = prot_dict.pop("protein")
        prot_dict["protein_descr"] = [
            p if p is not None else "" for p in prot_dict["protein_descr"]
        ]
        info.update(prot_dict)

        info.update(sh.pop("search_score"))
        mods = sh.pop("modifications", [])
        info["modifications"] = list(fix_mods(mods))

        for k, v in sh.items():
            if isinstance(v, (str, int, float)):
                info[k] = v
        if "analysis_result" in sh:
            for ar in sh["analysis_result"]:
                if ar["analysis"] == "peptideprophet":
                    result = ar["peptideprophet_result"]
                    try:
                        info.update(result["parameter"])
                    except KeyError:
                        pass
                    info["peptideprophet_probability"] = result["probability"]
                    info["peptideprophet_ntt_prob"] = result["all_ntt_prob"]
                elif ar["analysis"] == "interprophet":
                    result = ar["interprophet_result"]
                    info.update(result["parameter"])
                    info["interprophet_probability"] = result["probability"]
                    info["interprophet_ntt_prob"] = result["all_ntt_prob"]

    info["run"] = chop_spectrum(info["spectrum"])
    # is_rejected: Potential use in future for user manual validation (0 or 1)
    if "is_rejected" in info:
        info.pop("is_rejected")
    if "xcorr" in info:
        info["searchEngineScore"] = info["xcorr"]  # COMET
    elif "ionscore" in info:
        info["searchEngineScore"] = info["ionscore"]  # MASCOT

    return info


def pepxml_chunk(
    chunk: Iterator[dict[str, Any]],
    **kwargs: Any,
) -> pd.DataFrame | None:
    lst = [dict_to_series(r) for r in chunk]
    if not lst:
        return None
    return pd.DataFrame(lst, **kwargs)


def PepXMLDataFrame(
    filename: Path,
    stop: int | None = None,
    start: int = 0,
    step: int = 1,
    level: int = 0,
    number_of_bg_processes: int = 0,
    **kwargs: Any,
) -> pd.DataFrame:
    from .config import PEPXML_CHUNKS

    pep = Path(filename)

    it = PepXML(str(filename), use_index=True)  # need to get total
    total = len(it)
    if stop is not None or start != 0 or step != 1:
        # find size
        if stop is None:
            stop = total
        else:
            stop = min(stop, total)
        r = range(start, stop, step)
        total = len(r)
        it = islice(it, r.start, r.stop, r.step)
    ret = []
    it = log_iterator(
        it,
        total=total,
        level=level,
        desc=f"{pep.name}",
        number_of_bg_processes=number_of_bg_processes,
    )
    for chunk in ichunked(it, PEPXML_CHUNKS):
        df = pepxml_chunk(chunk, **kwargs)
        if df is None:
            continue
        ret.append(df)

    if not ret:
        raise RuntimeError(f"{pep.name}: no runs!")

    return pd.concat(ret, axis=0, ignore_index=True)


# at least version 2.5.2
class PyMzMLReader(Reader):
    pass
    # def _open_file(
    #     self, path_or_file: str, build_index_from_scratch=False
    # ) -> FileInterface:
    #     # sigh! after the iterator is exhausted the
    #     # reader reintializes the file... meaning that
    #     # indexes will be recreated for no reason
    #     if "file_object" in self.info and self.path_or_file is path_or_file:
    #         o: FileInterface = self.info["file_object"]
    #         # there is no api here just duck typing
    #         o.file_handler.file_handler.seek(0, 0)
    #         return o

    #     return super()._open_file(
    #         path_or_file, build_index_from_scratch=build_index_from_scratch
    #     )


def ProtXmlDataFrame(
    protxml: Path,
    level: int = 0,
    number_of_bg_processes: int = 0,
) -> pd.DataFrame:
    from pyteomics.protxml import ProtXML
    from .logger import log_iterator

    prot = ProtXML(str(protxml))
    total = len(prot)

    def gen_items(prot: ProtXML) -> Iterator[dict[str, Any]]:
        with prot as f:
            for item in f:
                info = {}
                for k, v in item.items():
                    if isinstance(v, (str, int, float)):
                        info[k] = v
                if "protein" in item:
                    for prot in item["protein"]:
                        out = dict(info)
                        out.update(prot)
                        if "indistinguishable_protein" in out:
                            out["indistinguishable_protein"] = [
                                p["protein_name"]
                                for p in out["indistinguishable_protein"]
                            ]
                        else:
                            out["indistinguishable_protein"] = []
                        yield out

    it = log_iterator(
        gen_items(prot),
        total=total,
        level=level,
        desc=f"{protxml.name}",
        number_of_bg_processes=number_of_bg_processes,
    )

    return pd.DataFrame(it)



from __future__ import annotations

import click

from .cli import cli
from .cli import CONFIG
from .cli import Config
from .cli import pass_config


def web_options(f):
    f = click.option(
        "-n",
        "--no-browse",
        is_flag=True,
        help="don't open web application in browser",
    )(f)
    f = click.option("-g", "--gunicorn", is_flag=True, help="run webapp with gunicorn")(
        f,
    )
    f = click.option(
        "--port",
        default=8000,
        help="port to run webservice on",
        show_default=True,
    )(f)
    return f


@cli.command(
    epilog="""

Running a webserver + turnover backend.

If not specifed with options or in a configuration file.
job results will be stored in `~/turnover_jobs` (created if not existing)
and file caches will use `~/turnover_cache` (created in not existing).

\b
Configuration
=============

If a configuration file is specified then you can specifiy
JOBSDIR, CACHEDIR.

The program will look for a file `~/.turnover.toml` for website
configuration if no `--web-config` is given.

Run `turnover web` and browse to "Configuration" for more information.

You can give arguments to the underlying webserver with arguments after
a `--`, e.g.:

`turnover web -g --workers=4 -- --workers=2`

will tell gunicorn to fire up 2 workers and use 4 background workers for turnover itself.
""",
)
@click.option(
    "-w",
    "--workers",
    type=int,
    help="number of background workers [default: half the number of cpus]",
)
@web_options
@click.option(
    "-c",
    "--web-config",
    type=click.Path(file_okay=True, dir_okay=False, exists=True),
    help="configuration file for web browser",
)
@click.option(
    "--jobs-dir",
    type=click.Path(file_okay=False),
    help="directory to run job [default: directory of ~/turnover_jobs]",
)
@click.option(
    "--cache-dir",
    type=click.Path(file_okay=False),
    help="directory to store file caches [default: directory of ~/turnover_cache]",
)
@click.option(
    "-v",
    "--view-only",
    is_flag=True,
    help="only view data",
)
@click.option(
    "--email",
    is_flag=True,
    help="allow background jobs to email user when done (requires a MAIL_SERVER configured)",
)
@click.argument("server_options", nargs=-1)
@pass_config
def web(
    cfg: Config,
    no_browse: bool,
    workers: int,
    web_config: str | None,
    gunicorn: bool = False,
    jobs_dir: str | None = None,
    cache_dir: str | None = None,
    port: int = 8000,
    server_options: tuple[str, ...] = (),
    view_only: bool = False,
    email: bool = False,
) -> None:
    """Run full website (requires protein-turnover-website)."""
    from os import cpu_count
    from .web import webrunner

    if workers is None:
        workers = max((cpu_count() or 1) // 2, 1)
    defaults = {}
    if cache_dir is not None:
        defaults["CACHEDIR"] = cache_dir

    if jobs_dir is not None:
        defaults["JOBSDIR"] = jobs_dir
    webrunner(
        browse=not no_browse,
        workers=workers,
        web_config=web_config,
        gunicorn=gunicorn,
        configfile=cfg.user_config,
        defaults=defaults,
        port=port,
        server_options=server_options,
        view_only=view_only,
        email=email,
    )


@cli.command()
@web_options
@click.argument("jobfile", type=click.Path(file_okay=True, exists=True, dir_okay=False))
@click.argument("server_options", nargs=-1)
@pass_config
def view(
    cfg: Config,
    no_browse: bool,
    jobfile: str,
    gunicorn: bool = False,
    port: int = 8000,
    server_options: tuple[str, ...] = (),
) -> None:
    """View a completed run in a browser (requires protein-turnover-website)"""
    from pathlib import Path
    from .web import webrunner
    from .jobs import TurnoverJob
    from .utils import ResultsResourceFile

    jf = Path(jobfile).expanduser().resolve()
    if not jf.exists():
        click.secho(f"no such file {jobfile}", fg="red")
        raise click.Abort()

    job = TurnoverJob.restore(jf)
    jobsddir = jf.parent
    if job.auto_generated_jobid:
        results = ResultsResourceFile(job.jobid, jobsddir)
        if not results.has_result():
            raise ValueError(f"can't find output data file {results.cache_result()}")
    dataid = jf.stem
    defaults = {"JOBSDIR": str(jobsddir), "DATAID": dataid}

    webrunner(
        browse=not no_browse,
        workers=1,
        web_config=None,
        gunicorn=gunicorn,
        configfile=cfg.user_config,
        defaults=defaults,
        port=port,
        view_only=True,
        flask_app="protein_turnover_website.wsgi_view",
        server_options=server_options,
    )


@cli.command(hidden=True)
@web_options
@click.argument("server_options", nargs=-1)
@pass_config
def config(
    cfg: Config,
    no_browse: bool,
    gunicorn: bool = False,
    port: int = 8000,
    server_options: tuple[str, ...] = (),
) -> None:
    """View Configuration information"""
    from .web import webrunner

    webrunner(
        browse=not no_browse,
        workers=1,
        web_config=None,
        defaults={"DATAID": "noid"},
        gunicorn=gunicorn,
        port=port,
        # extra=web_options,
        view_only=True,
        flask_app="protein_turnover_website.wsgi_view",
        page="configuration.html",
        server_options=server_options,
    )


@cli.command()
@click.option(
    "-o",
    "--out",
    type=click.Path(dir_okay=False),
    help="write configuration file here [default ~/.turnover.toml]",
)
@click.option("--force", is_flag=True, help="force overwriting of any existing file")
@click.option("--with-website", is_flag=True, help="add website configuration too")
def init_config(force: bool, with_website: bool, out: str | None):
    """Create a default configuration file"""
    from pathlib import Path

    from .web import has_package, dump_config

    if out is None:
        configfile = CONFIG
    else:
        configfile = Path(out)

    if configfile.exists() and not force:
        click.secho(f"{configfile} file exists! Use --force to overwrite.")
        return
    if with_website and not has_package("protein_turnover_website"):
        click.secho(
            "Please install protein_turnover_website [pip install protein-turnover-website]!",
            fg="red",
            err=True,
        )
        raise click.Abort()

    dump_config(configfile, with_website)



from __future__ import annotations

from dataclasses import dataclass
from typing import TypedDict

import numpy as np

# generated by df2class


@dataclass
class TurnoverRow:
    fraction: str
    peptide: str
    run: str
    mz: float
    theoreticalMass: float
    observedMass: float
    charge: int
    retentionTime: float
    missedCleavages: int
    proteins: np.ndarray  # item: str
    modifications: np.ndarray
    peptideprophet_probability: float
    probability: float
    searchEngineScore: float
    searchEngineName: str
    description: str
    isLabelledRun: bool
    maxIso: int
    labelledElementCount: int
    # isotopeRegressions: np.ndarray  # item: float32
    adjustedRsq: np.ndarray  # item: float32
    isotopeEnvelopes: np.ndarray  # item: float64
    monoFitParams: np.ndarray  # item: float32

    # heavyDistribution: np.ndarray  # item: float64
    # heavyFitParams: np.ndarray  # item: float32

    heavyCor: float
    labelledEnvelopes: np.ndarray  # item: float64
    relativeIsotopeAbundance: float
    enrichment: float
    theoreticalDist: np.ndarray
    # idx: int
    prStr: str
    prGroup: int
    eics_idx: str
    eics: np.ndarray  # item: float32
    mzranges: np.ndarray  # item: float32
    eics_shape: np.ndarray  # item: int64
    modcol: str
    fdr: float
    maxPeakArea: float


class TurnoverDict(TypedDict):
    fraction: str
    peptide: str
    run: str
    mz: float
    theoreticalMass: float
    observedMass: float
    charge: int
    retentionTime: float
    missedCleavages: int
    proteins: np.ndarray  # item: str
    modifications: np.ndarray
    peptideprophet_probability: float
    searchEngineScore: float
    searchEngineName: str
    description: str
    isLabelledRun: bool
    maxIso: int
    labelledElementCount: int
    # isotopeRegressions: np.ndarray  # item: float32
    adjustedRsq: np.ndarray  # item: float32
    isotopeEnvelopes: np.ndarray  # item: float64
    monoFitParams: np.ndarray  # item: float32
    # heavyDistribution: np.ndarray  # item: float64
    # heavyFitParams: np.ndarray  # item: float32
    labelledEnvelopes: np.ndarray  # item: float64
    relativeIsotopeAbundance: float

    enrichment: float
    labelEnrichment2: float
    heavyCor: float
    heavyCor2: float
    totalNNLSWeight: float
    nnls_residual: float
    theoreticalDist: np.ndarray  # item: float32
    # idx: int
    prStr: str
    prGroup: int
    eics_idx: str
    eics: np.ndarray  # item: float32
    mzranges: np.ndarray  # item: float32
    eics_shape: np.ndarray  # item: int64
    modcol: str
    fdr: float
